from __future__ import annotations

import abc
import collections
import uuid
import numpy as np
import os
from .. import Devutils as dev
from ..Jupyter import JHTML, X3DHTML
from .. import Numputils as nput

__all__ = [
    "X3D",
    "X3DPrimitive",
    "X3DGeometryObject",
    "X3DGeometryGroup",
    "X3DGroup",
    "X3DScene",
    "X3DBackground",
    "X3DMaterial",
    "X3DLine",
    "X3DSphere",
    "X3DCone",
    "X3DCylinder",
    "X3DCappedCylinder",
    "X3DArrow",
    "X3DTorus",
    "X3DRectangle2D",
    "X3DDisk2D",
    "X3DCircle2D",
    "X3DPolyline2D",
    "X3DTriangleSet",
    "X3DIndexedTriangleSet",
    "X3DIndexedLineSet",
    "X3DSwitch",
    "X3DListAnimator",
    "X3DInterpolatingAnimator"
]

#TODO: cache these resources or put them on a path that is accessible by Jupyter
#      might be pretty simple depending on what resource paths Jupyter naturally exposes

class X3DObject(metaclass=abc.ABCMeta):
    id: str
    @abc.abstractmethod
    def to_x3d(self) -> X3DHTML.X3DElement:
        ...

    def get_interpolated_attributes(self):
        return self.to_x3d().attrs
    def get_children(self):
        return []

    @classmethod
    def get_new_id(cls):
        return str(uuid.uuid4())[:6]

    def resolve_prop_attr(self, prop_name):
        return prop_name
    def get_prop_node_id(self, prop_name):
        return self.id

class X3D(X3DObject):
    defaults = dict(
        width=500,
        height=500
    )
    @classmethod
    def get_new_id(cls):
        return "x3d-" + str(uuid.uuid4())[:6]
    def __init__(self, *children, id=None, dynamic_loading=True, x3dom_path=None, x3dom_css_path=None, **opts):
        if len(children) == 1 and isinstance(children[0], (tuple, list)):
            children = children[0]
        self.children = children
        self.opts = opts
        if id is None:
            id = self.get_new_id()
        self.id = id
        self.dynamic_loading = dynamic_loading
        if x3dom_path is not None:
            if dev.str_is(x3dom_path, 'local') and not os.path.isfile('local'):
                # get the relative path
                x3dom_path = 'file://' + os.path.join(
                    os.path.dirname(os.path.dirname(__file__)),
                    'Jupyter', 'resources', 'x3dom-full.js'
                )
            self.X3DOM_JS = x3dom_path
        if x3dom_css_path is not None:
            self.X3DOM_CSS = x3dom_css_path

        self._widg = None

    X3DOM_JS = 'https://www.x3dom.org/download/1.8.3/x3dom-full.js'
    X3DOM_CSS = 'https://www.x3dom.org/download/x3dom.css'

    @classmethod
    def get_export_script(self, id):
        return f"""
(function(){{
  let link = document.createElement('a');
  let base_name = '{id}';
  link.download = base_name + '.png';
  link.href = document.getElementById('{id}').getElementsByTagName('canvas')[0].toDataURL()
  link.click();
}})()
       """

    @classmethod
    def get_view_settings_script(self, id):
        return f"""
    (function(){{
        let fig = document.getElementById('{id}');
        let x3d = fig.getElementsByTagName("x3d")[0];
        let vmObj = x3d.runtime.viewMatrix();
        let vmA = [
            [vmObj["_00"], vmObj["_01"], vmObj["_02"], vmObj["_03"]],
            [vmObj["_10"], vmObj["_11"], vmObj["_12"], vmObj["_13"]],
            [vmObj["_20"], vmObj["_21"], vmObj["_22"], vmObj["_23"]],
            [vmObj["_30"], vmObj["_31"], vmObj["_32"], vmObj["_33"]]
        ];
        let out = document.getElementById('{id}-view-matrix');
        out.value = JSON.stringify(vmA, 1);
    }})()
           """

    @classmethod
    def parse_view_matrix(cls, vs):
        import json
        if isinstance(vs, str):
            vs = json.loads(vs)
        vm = np.linalg.inv(vs)
        ang, ax = nput.extract_rotation_angle_axis(vm[:3, :3])
        v_pos = vm[:3, -1].tolist()
        v_ort = np.array(list(ax) + [ang]).tolist()
        opts = {"position": v_pos, "orientation": v_ort}
        return opts

    @classmethod
    def get_record_screen_script(self, id, polling_rate=30, recording_duration=2, video_format='video/webm'):
            return f"""
    (function(){{
        let canvas = document.getElementById('{id}').getElementsByTagName('canvas')[0];
        
        let pollingRate = (typeof canvas.pollingRate === 'undefined') ? {polling_rate} : canvas.pollingRate;
        let videoFormat = (typeof canvas.videoFormat === 'undefined') ? "{video_format}" : canvas.videoFormat;
        let videoExtension = canvas.videoExtension;
        if (typeof canvas.videoExtension === 'undefined') {{
            videoExtension = ""
        }}
        let x3DRecordingStream = canvas.captureStream(pollingRate);
        let mediaRecorder = new MediaRecorder(x3DRecordingStream, {{mimeType: videoFormat}});
        
        mediaRecorder.frames = [];
        mediaRecorder.ondataavailable = function(e) {{
          mediaRecorder.frames.push(e.data);
        }};
        
        mediaRecorder.onstop = function(e) {{
          link = document.createElement('a');
          const base_name = '{id}';
          const blob = mediaRecorder.frames[0];
          link.download = base_name + videoExtension;
          console.log(blob);
          const blobURL = window.URL.createObjectURL(blob);
          link.href = blobURL;
          console.log(blobURL);
          mediaRecorder.frames = [];
          link.click();
        }};
        
        let duration = (typeof canvas.recordingDuration === 'undefined') ? {recording_duration} : canvas.recordingDuration;
        setTimeout(() => {{mediaRecorder.stop()}}, duration * 1000);
        mediaRecorder.start()
    }})()
           """
    @classmethod
    def set_animation_duration_script(self, id):
        return f"""
    (function(){{
        let canvas = document.getElementById('{id}').getElementsByTagName('canvas')[0];
        let input = document.getElementById('{id}-duration-input');
        
        canvas.recordingDuration = input.value;
    }})()
           """

    include_export_button = False
    include_record_button = False
    include_view_settings_button = False
    def to_widget(self, dynamic_loading=None,
                  include_export_button=None,
                  include_record_button=None,
                  include_view_settings_button=None
                  ):
        if self._widg is not None:
            return self._widg
        id = self.id
        x3d_embed = self.to_x3d()#.tostring()

        if dynamic_loading is None:
            dynamic_loading = self.dynamic_loading
        if include_export_button is None:
            include_export_button = self.include_export_button
        if include_record_button is None:
            include_record_button = self.include_record_button
        if include_view_settings_button is None:
            include_view_settings_button = self.include_view_settings_button

        if not dynamic_loading:
            base_fig = JHTML.Div(
                JHTML.Link(rel='stylesheet', href=self.X3DOM_CSS),
                JHTML.Script(src=self.X3DOM_JS),
                x3d_embed,
                id=id,
                width=x3d_embed['width'],
                height=x3d_embed['height'],
                can_be_dynamic=False
            )
        else:
            JHTML.Link(rel='stylesheet', href=self.X3DOM_CSS),
            load_script = JHTML.Script(src=self.X3DOM_JS).tostring()
            kill_id = "tmp-"+str(uuid.uuid4())[:10]
            base_fig = JHTML.Figure(
                # JHTML.Link(rel='stylesheet', href=self.X3DOM_CSS),
                x3d_embed,
                JHTML.Image(
                    src='data:image/gif;base64,R0lGODlhAQABAIAAAAAAAP///yH5BAEAAAAALAAAAAABAAEAAAIBRAA7',
                    id=kill_id,
                    onload=f"""
                    (function() {{
                        let killElem = document.getElementById("{kill_id}");
                        if (killElem !== null) {{
                            killElem.remove();
                            const frag = document.createRange().createContextualFragment(`{load_script}`);
                            document.head.appendChild(frag);
                        }}
                    }})()"""
                    ),
                id=id,
                width=x3d_embed['width'],
                height=x3d_embed['height'],
                can_be_dynamic=False
            )

        elems = [base_fig]
        if include_export_button:
            elems.append(JHTML.Button("Save Figure", onclick=self.get_export_script(self.id)))
        if include_record_button:
            elems.extend([
                JHTML.Button("Record Animation", onclick=self.get_record_screen_script(self.id)),
                JHTML.Input(value="2", id=self.id+'-duration-input', width="50px", oninput=self.set_animation_duration_script(self.id))
            ])
        if include_view_settings_button:
            elems.append(
                JHTML.Div(
                    [
                        JHTML.Button("Show View Matrix", onclick=self.get_view_settings_script(self.id)),
                        JHTML.Textarea(id=self.id + '-view-matrix')
                    ],
                    display="block"
                )
            )

        if len(elems) > 1:
            self._widg = JHTML.Div(
                *elems
            )
        else:
            self._widg = elems[0]

        return self._widg

    def to_html(self, *base_elems, header_elems=None,
                dynamic_loading=False,
                include_export_button=None,
                include_record_button=None,
                **header_info):
        id = self.id
        x3d_embed = self.to_widget(
            dynamic_loading=dynamic_loading,
            include_export_button=include_export_button,
            include_record_button=include_record_button
        )  # .tostring()

        return JHTML.Html(
            JHTML.Head(
                *(header_elems if header_elems is not None else []),
                JHTML.Link(rel='stylesheet', href=self.X3DOM_CSS),
                JHTML.Script(src=self.X3DOM_JS),
                **header_info
            ),
            JHTML.Body(
                *base_elems,
                x3d_embed
            ),
            id=id
        )

    def _ipython_display_(self):
        return self.to_widget()._ipython_display_()
    def get_mime_bundle(self):
        return self.to_widget().get_mime_bundle()
    def to_x3d(self):
        base_opts = dict(self.defaults, **self.opts)
        for k in ['width', 'height']:
            if k in base_opts:
                v = base_opts[k]
                if nput.is_numeric(v):
                    base_opts[k] = f'{v:.0f}px'
        return X3DHTML.X3D(
            JHTML.HTML.Head(),
            *[a.to_x3d() if hasattr(a, 'to_x3d') else a for a in self.children],
            **base_opts
        )
    def display(self):
        return self.to_widget().display()

    def show(self):
        from ..Jupyter.JHTML import JupyterAPIs
        dynamic_loading = JupyterAPIs().in_jupyter_environment()
        self.to_widget(dynamic_loading=dynamic_loading).display()

    def dump(self, file, write_html=True, **opts):
        if write_html:
            html = self.to_html()
        else:
            html = self.to_x3d()
        return html.write(file, **opts)

    def get_children(self):
        return self.children

class X3DOptionsSet(X3DObject):
    __props__ = {}

    @classmethod
    def get_new_id(cls):
        return "x3d-opts-" + str(uuid.uuid4())[:6]
    def __init__(self, id=None, **attrs):
        if id is None:
            id = self.get_new_id()
        self.id = id
        self.attrs = attrs

    conversion_map = {}
    @classmethod
    def prop_keys(cls):
        return (cls.__props__ | cls.conversion_map.keys())
    def prep_attrs(self, attrs:dict):
        attrs = {
            self.conversion_map.get(k, k):v
            for k,v in attrs.items()
        }
        excess_keys = attrs.keys() - self.__props__
        if len(excess_keys) > 0:
            cls = type(self).__name__
            raise ValueError(f"keys {excess_keys} are invalid keys for {cls}")
        attrs['id'] = self.id
        return attrs

    @classmethod
    def resolve_prop_attr(self, prop_name):
        return self.conversion_map.get(prop_name, prop_name)

    def get_prop_node_id(self, prop_name):
        return self.id

class X3DMaterial(X3DOptionsSet):
    __props__ = {
        "diffuseColor",
        "ambientIntensity",
        "emissiveColor",
        "specularColor",
        "shininess",
        "transparency",
        # "metadata"
    }
    conversion_map = {
        "brightness": "ambientIntensity",
        "glow": "emissiveColor",
        "color": "diffuseColor",
        "specularity": "specularColor"
    }
    def to_x3d(self):
        return X3DHTML.Material(**self.prep_attrs(self.attrs))

class X3DAppearance(X3DOptionsSet):
    __props__ = {
        "alphaClipThreshold",
        "blendMode",
        "colorMaskMode"
        "depthMode"
        "lineProperties"
        "material"
        "metadata"
        "pointProperties"
        "shaders",
        "sortKey",
        "sortType",
        "texture",
        "textureTransform"
    }
    @classmethod
    def get_new_id(cls):
        return "x3d-appearance-" + str(uuid.uuid4())[:6]
    def prep_attrs(self, attrs:dict):
        material_keys = attrs.keys() & X3DMaterial.prop_keys()
        line_keys = attrs.keys() & X3DLineProperties.prop_keys()
        point_keys = attrs.keys() & X3DPointProperties.prop_keys()

        rem_keys = attrs.keys() - (
            X3DMaterial.prop_keys()
            | X3DLineProperties.prop_keys()
            | X3DPointProperties.prop_keys()
        )
        base_attrs = {k:attrs[k] for k in rem_keys}

        if len(material_keys) > 0:
            base_attrs['material'] = {k:attrs[k] for k in material_keys}

        if len(line_keys) > 0:
            base_attrs['lineProperties'] = {k:attrs[k] for k in line_keys}

        if len(point_keys) > 0:
            base_attrs['pointProperties'] = {k:attrs[k] for k in point_keys}

        return base_attrs
    def to_x3d(self):
        base_attrs = self.prep_attrs(self.attrs)

        comps = []
        line_props = base_attrs.pop('lineProperties', None)
        if line_props is not None:
            if isinstance(line_props, dict):
                line_props = X3DLineProperties(id=self.id+'-lineprops', **line_props)
            if isinstance(line_props, X3DLineProperties):
                line_props = line_props.to_x3d()
            comps.append(line_props)

        point_props = base_attrs.pop('pointProperties', None)
        if point_props is not None:
            if isinstance(point_props, dict):
                point_props = X3DPointProperties(id=self.id+'-pointprops', **point_props)
            if isinstance(point_props, X3DPointProperties):
                point_props = point_props.to_x3d()
            comps.append(point_props)

        material_props = base_attrs.pop('material', None)
        if material_props is not None:
            if isinstance(material_props, dict):
                material_props = X3DMaterial(id=self.id+'-material', **material_props)
            if isinstance(material_props, X3DMaterial):
                material_props = material_props.to_x3d()
            comps.append(material_props)

        return X3DHTML.Appearance(*comps, **base_attrs)

    @classmethod
    def resolve_prop_attr(self, prop_name):
        if prop_name in X3DAppearance.prop_keys():
            return self.conversion_map.get(prop_name, prop_name)
        elif prop_name in X3DLineProperties.prop_keys():
            return X3DLineProperties.conversion_map.get(prop_name, prop_name)
        elif prop_name in X3DLineProperties.prop_keys():
            return X3DLineProperties.conversion_map.get(prop_name, prop_name)
        elif prop_name in X3DMaterial.prop_keys():
            return X3DMaterial.conversion_map.get(prop_name, prop_name)
        else:
            raise ValueError(f"property {prop_name} not known")

    def get_prop_node_id(self, prop_name):
        if prop_name in X3DAppearance.prop_keys():
            return self.id
        elif prop_name in X3DLineProperties.prop_keys():
            return self.id+'-lineprops'
        elif prop_name in X3DLineProperties.prop_keys():
            return self.id+'-pointprops'
        elif prop_name in X3DMaterial.prop_keys():
            return self.id+'-material'
        else:
            raise ValueError(f"property {prop_name} not known")

class X3DLineProperties(X3DOptionsSet):
    __props__ = {
        "applied",
        "linetype",
        "linewidth",
        "linewidthScaleFactor"
    }
    conversion_map = {
        "line_style":"linetype",
        "line_thickness":"linewidth"
    }
    def prep_attrs(self, attrs:dict):
        attrs = super().prep_attrs(attrs)
        attrs['linewidthScaleFactor'] = attrs.get('linewidthScaleFactor', '1')
        attrs['containerField'] = attrs.get('containerField', 'lineProperties')
        return attrs
    def to_x3d(self):
        return X3DHTML.LineProperties(**self.prep_attrs(self.attrs))

class X3DPointProperties(X3DOptionsSet):
    __props__ = {
        "attenuation",
        "pointSizeMaxValue",
        "pointSizeMinValue",
        "pointSizeScaleFactor"
    }
    conversion_map = {
        "point_size":"pointSizeScaleFactor"
    }
    def prep_attrs(self, attrs:dict):
        attrs = super().prep_attrs(attrs)
        attrs['containerField'] = attrs.get('containerField', 'pointProperties')
        attrs['pointSizeMaxValue'] = str(
            max([
                float(attrs.get('pointSizeMaxValue', '0')),
                float(attrs.get('pointSizeMinValue', '0')),
                float(attrs.get('pointSizeScaleFactor', '0')),
            ])
        )
        return attrs
    def to_x3d(self):
        return X3DHTML.PointProperties(**self.prep_attrs(self.attrs))

class X3DPrimitive(X3DObject):
    wrapper_class = None
    tag_class = None
    @classmethod
    def get_new_id(cls):
        return "x3d-obj-" + str(uuid.uuid4())[:6]
    def __init__(self, *children, id=None, **opts):
        if len(children) == 1 and isinstance(children[0], (tuple, list)):
            children = children[0]
        self.children = children
        if id is None:
            id = self.get_new_id()
        opts['id'] = id
        self.opts = opts
    @property
    def id(self):
        return self.opts['id']
    @id.setter
    def id(self, new_id):
        self.opts['id'] = new_id
    def split_opts(self, opts:dict):
        material_keys = opts.keys() & (
            X3DMaterial.prop_keys()
            | X3DAppearance.prop_keys()
            | X3DLineProperties.prop_keys()
            | X3DPointProperties.prop_keys()
        )
        rem_keys = opts.keys() - material_keys
        return {k:opts[k] for k in rem_keys}, {k:opts[k] for k in material_keys}
    def get_appearance(self, appearance_options):
        if len(appearance_options) > 0:
            return X3DAppearance(id=self.id+"-appearance", **appearance_options).to_x3d()
        else:
            return None
    def to_x3d(self):
        obj_opts, appearance_opts = self.split_opts(self.opts)
        kids = [k.to_x3d() if hasattr(k, 'to_x3d') else k for k in self.children]
        appearance = self.get_appearance(appearance_opts)
        if self.tag_class is None:
            if appearance is not None:
                kids = [appearance] + kids
            return self.wrapper_class(
                kids,
                **obj_opts
            )
        else:
            core = self.tag_class(kids, **obj_opts)
            appearance = self.get_appearance(appearance_opts)
            if appearance is not None:
                return self.wrapper_class(appearance, core)
            else:
                return core

    @classmethod
    def resolve_prop_attr(self, prop_name):
        if prop_name in (
                X3DMaterial.prop_keys()
                | X3DAppearance.prop_keys()
                | X3DLineProperties.prop_keys()
                | X3DPointProperties.prop_keys()
        ):
            return X3DAppearance.resolve_prop_attr(prop_name)
        #TODO: handle mapping of transform props
        else:
            return prop_name

    def get_prop_node_id(self, prop_name):
        if prop_name in (
            X3DMaterial.prop_keys()
            | X3DAppearance.prop_keys()
            | X3DLineProperties.prop_keys()
            | X3DPointProperties.prop_keys()
        ):
            return X3DAppearance(id=self.id + "-appearance").get_prop_node_id(prop_name)
        else:
            return self.id

    def get_children(self):
        return self.children

class X3DScene(X3DPrimitive):
    wrapper_class = X3DHTML.Scene
    default_viewpoint = {'viewAll':True}
    children: list
    def __init__(self, *children:X3DPrimitive, background=None, viewpoint=None, **opts):
        if viewpoint is None:
            viewpoint = self.default_viewpoint
        elif viewpoint is False:
            viewpoint = {}
        super().__init__(*children, **opts)
        if background is not None:
            self.children = [X3DBackground(color=background)] + list(self.children)
        if len(viewpoint) > 0:
            viewpoint = self.get_view_settings(**viewpoint)
            self.children = [X3DHTML.Viewpoint(**viewpoint)] + list(self.children)

    default_up_vector = (0, 1, 0)
    default_right_vector = (1, 0, 0)
    default_view_vector = (0, 0, 1)
    default_view_distance = 10
    @classmethod
    def get_view_settings(cls,
                          up_vector=None, view_vector=None, right_vector=None,
                          view_distance=None,
                          view_center=None,
                          view_matrix=None,
                          view_position=None,
                          **etc):
        # CO = coords0[1] - coords0[0]
        # OH = coords0[5] - coords0[1]
        if view_matrix is None:
            if view_vector is None:
                if (
                    up_vector is not None and right_vector is not None
                ):
                    view_vector = nput.vec_crosses(up_vector, right_vector, normalize=True)
                elif right_vector is not None:
                    view_vector = nput.vec_crosses(cls.default_up_vector, right_vector, normalize=True)
                elif up_vector is not None:
                    view_vector = nput.vec_crosses(up_vector, cls.default_right_vector, normalize=True)

            if view_vector is not None:
                m = nput.rotation_matrix(
                    view_vector,
                    cls.default_view_vector
                )
            else:
                m = np.eye(3)

            if up_vector is None and right_vector is not None:
                if view_vector is None:
                    view_vector = cls.default_view_vector
                up_vector = nput.vec_normalize(
                    nput.vec_crosses(right_vector, view_vector)
                )
            elif up_vector is not None and view_vector is not None:
                up_vector = nput.vec_crosses(
                    view_vector,
                    nput.vec_crosses(view_vector, up_vector),
                    normalize=True
                )
            if up_vector is not None:
                m = m @ nput.rotation_matrix(
                    m.T @ up_vector,
                    cls.default_up_vector
                )
            view_matrix = m

        ang, cross = nput.extract_rotation_angle_axis(view_matrix)
        if view_vector is None:
            view_vector = view_matrix[:, -1]
        if view_position is None:
            if view_distance is None:
                view_distance = cls.default_view_distance
            view_position = view_distance * nput.vec_normalize(np.asanyarray(view_vector))
            if view_center is not None:
                if isinstance(view_center, dict):
                    view_center = view_center['untransformed']
                else:
                    view_center = view_matrix @ np.asanyarray(view_center)
                view_position = view_distance * nput.vec_normalize(
                    view_position + view_center
                )
        else:
            if isinstance(view_position, dict):
                view_position = view_position['untransformed']
            else:
                view_position = view_matrix @ np.asanyarray(view_position)
        return dict(
            {
                'orientation': list(cross) + [ang],
                'position': view_position
            },
            **etc
        )

class X3DBackground(X3DOptionsSet):
    wrapper_class = X3DHTML.Background
    __props__ = {
        'skyColor',
        'skyAngle'
    }
    conversion_map = {
        "color": "skyColor"
    }
    def prep_attrs(self, attrs: dict):
        attrs = super().prep_attrs(attrs)
        color = attrs.get('skyColor', None)
        if color is not None:
            if isinstance(color, (list, tuple, np.ndarray)) and all(isinstance(c, str) for c in color):
                color = " ".join(color)
            if isinstance(color, str):
                try:
                    _ = [float(s) for s in color.split()]
                except ValueError:
                    from .Colors import ColorPalette

                    bits = []
                    for c in color.split():
                        if color.startswith('#'):
                            c = ColorPalette.parse_rgb_code(c)
                        else:
                            c = np.array(ColorPalette.parse_color_string(c))
                        bits.extend(np.array(c) / 255)
                    attrs['skyColor'] = bits
        return attrs

    def to_x3d(self):
        return X3DHTML.Background(**self.prep_attrs(self.attrs))

class X3DCoordinate(X3DPrimitive):
    wrapper_class = X3DHTML.Coordinate
    def __init__(self, points):
        super().__init__(point=self.prep_points(points))
    @classmethod
    def prep_points(cls, points):
        return " ".join(np.asanyarray(np.round(points, 4)).flatten().astype(str))

class X3DColor(X3DPrimitive):
    wrapper_class = X3DHTML.Color
    def __init__(self, colors):
        super().__init__(color_list=colors)
    def split_opts(self, opts:dict):
        base_opts, appearance_opts = super().split_opts(opts)
        base_opts['color'] = self.prep_color(base_opts.pop('color_list'))
        return base_opts, appearance_opts
    @classmethod
    def prep_color(cls, points):
        return " ".join(np.asanyarray(np.round(points, 4)).flatten().astype(str))

class X3DGroup(X3DPrimitive):
    wrapper_class = X3DHTML.Group

class X3DSwitch(X3DPrimitive):
    wrapper_class = X3DHTML.Switch

class X3DGeometryObject(X3DPrimitive):
    wrapper_class = X3DHTML.Shape
    def __init__(self, *args, **opts):
        geom_opts, self.material_opts = self.split_opts(opts)
        self.geometry_opts = self.prep_geometry_opts(*args, **geom_opts)
        super().__init__()
    def get_interpolated_attributes(self):
        return dict(self.geometry_opts, **self.material_opts)
    @abc.abstractmethod
    def prep_geometry_opts(self, *args, **opts) -> dict:
        ...
    def create_tag_object(self, **core_opts):
        return self.tag_class(**core_opts)
    def create_object(self,
                      translation=None,
                      rotation=None,
                      scale=None,
                      normal=None,
                      up_vector=None,
                      bbox_center=None,
                      **core_opts):
        core_opts['id'] = core_opts.get('id', self.id)
        base_obj = self.create_tag_object(**core_opts)
        tf = {}
        if normal is not None:
            if up_vector is None:
                up_vector = [0, 0, 1]
            angs, crosses = nput.vec_angles(up_vector, normal, return_crosses=True, return_norms=False)
            if rotation is not None:
                if isinstance(rotation, str):
                    rotation = np.array(rotation.split()).astype(float)
                angs, crosses = nput.extract_rotation_angle_axis(
                    nput.rotation_matrix(crosses, angs)
                        @ nput.rotation_matrix(rotation[:3], rotation[3])
                )
            rotation = np.concatenate([crosses, [angs]])
        for k,v in [["translation",translation], ["rotation",rotation], ["scale", scale], ['bboxcenter', bbox_center]]:
            if v is not None:
                tf[k] = np.round(v, 4) if not isinstance(v, str) else v
        if len(tf) == 0:
            tf = None
        # base_obj = X3DHTML.Transform(base_obj, translation=translation, rotation=rotation, scale=scale)
        return base_obj, tf
    def get_rotation(self, axis, up_vector=None):
        if up_vector is None:
            up_vector = [0, 1, 0]
        angs, crosses, norms = nput.vec_angles(up_vector, axis, return_crosses=True, return_norms=True)
        if nput.is_numeric(angs):
            return np.concatenate([crosses, [angs]]), norms[1]
        else:
            return np.concatenate([crosses, angs[..., np.newaxis]], axis=-1), norms[1]

    transform_props = ("translation", "rotation", "scale", "bboxcenter")
    def get_prop_node_id(self, prop_name):
        if prop_name in self.transform_props:
            return self.id + "-transform"
        else:
            return self.id
    def to_x3d(self):
        # obj_opts, material_opts = self.split_opts(self.opts)
        # kids = [k.to_x3d() for k in self.children]
        core, tf = self.create_object(**self.geometry_opts)
        appearance = self.get_appearance(self.material_opts)
        if appearance is not None:
            core = self.wrapper_class(appearance, core)
        if tf is not None:
            core = X3DHTML.Transform(core, id=core.id+"-transform", **tf)
        return core

class X3DGeometryGroup(X3DGeometryObject):
    @abc.abstractmethod
    def prep_geometry_opts(self, *args, **opts) -> list[dict]:
        ...
    def get_interpolated_attributes(self):
        return dict(self.geometry_opts[0], **self.material_opts)
    def prep_vecs(self, vecs, nstruct=None):
        if vecs is None:
            return [None] * nstruct
        else:
            vecs = np.asanyarray(vecs)
            if vecs.ndim == 1:
                vecs = vecs[np.newaxis]
            if nstruct is not None:
                vecs = np.broadcast_to(vecs, (nstruct, vecs.shape[-1]))
        return vecs
    def prep_mats(self, mats, nstruct=None):
        if mats is None:
            return [None] * nstruct
        else:
            mats = np.asanyarray(mats)
            if mats.ndim == 2:
                mats = mats[np.newaxis]
            if nstruct is not None:
                mats = np.broadcast_to(mats, (nstruct,) + mats.shape[-2:])
            return mats
    def prep_const(self, const, nstruct):
        if const is None:
            return [None] * nstruct
        else:
            const = np.asanyarray(const)
            if const.ndim == 0:
                const = const[np.newaxis]
            const = np.broadcast_to(const, (nstruct,))
            return const
    def to_x3d(self):
        kids = [self.create_object(**g) for g in self.geometry_opts]
        appearance = self.get_appearance(self.material_opts)
        objs = []
        for i,(o,tf) in enumerate(kids):
            if hasattr(o, 'id'):
                id = o.id
            else:
                try:
                    id = o['id']
                except KeyError:
                    id = None
            if appearance is not None:
                o = self.wrapper_class(appearance, o)
            if tf is not None:
                o = X3DHTML.Transform(o, id=id+"-transform", **tf)
            objs.append(o)
        if len(objs) == 1:
            return objs[0]
        else:
            return X3DHTML.Group(objs)

class X3DSphere(X3DGeometryGroup):
    tag_class = X3DHTML.Sphere

    def prep_geometry_opts(self, centers, radius=1, **opts):
        centers = self.prep_vecs(centers)
        rads = self.prep_const(radius, centers.shape[0])
        return [{"translation":c, "radius":r} for c,r in zip(centers, rads)]

class X3DCylinder(X3DGeometryGroup):
    tag_class = X3DHTML.Cylinder

    def prep_geometry_opts(self, starts, ends, radius=1, **opts):
        starts = self.prep_vecs(starts)
        ends = self.prep_vecs(ends)
        radius = self.prep_const(radius, starts.shape[0])

        axes = ends - starts
        rots, norms = self.get_rotation(axes)

        return [
            {"translation":s, "rotation":a, "height":n, "radius":r}
            for s,a,n,r in zip((starts + ends) / 2, rots, norms, radius)
        ]

class X3DCone(X3DGeometryGroup):
    tag_class = X3DHTML.Cone

    def prep_geometry_opts(self, starts, ends, radius=1, top_radius=None, **opts):
        starts = self.prep_vecs(starts)
        ends = self.prep_vecs(ends)
        radius = self.prep_const(radius, starts.shape[0])
        top_radius = self.prep_const(top_radius, starts.shape[0])

        axes = ends - starts
        rots, norms = self.get_rotation(axes)

        return [
            {"translation":s, "rotation":a, "height":n, "bottomRadius":r, "topRadius":t}
            for s,a,n,r,t in zip((starts + ends) / 2, rots, norms, radius, top_radius)
        ]

class X3DArrow(X3DGroup):

    arrowhead_class = X3DCone
    cylinder_class = X3DCylinder

    def __init__(self,
                 starts, ends,
                 radius=1,
                 top_radius=None,
                 arrowhead_radius=2,
                 arrowhead_radius_mode='scaled',
                 arrowhead_offset=.3,
                 arrowhead_offset_mode='scaled',
                 cylinder_class=None,
                 arrowhead_class=None,
                 **opts):
        if arrowhead_class is None:
            arrowhead_class = self.arrowhead_class
        if cylinder_class is None:
            cylinder_class = self.cylinder_class

        ends = np.asanyarray(ends)
        starts = np.asanyarray(starts)
        arrow_vectors = ends - starts
        norms = nput.vec_norms(arrow_vectors)
        if arrowhead_offset_mode == 'scaled':
            arrowhead_offset = arrowhead_offset * norms
        disp_vectors = arrowhead_offset * nput.vec_normalize(arrow_vectors, norms=norms)
        arrow_starts = ends - disp_vectors
        if arrowhead_radius_mode == 'scaled':
            arrowhead_radius = arrowhead_radius * radius
        arrowheads = arrowhead_class(arrow_starts, ends,
                                     radius=arrowhead_radius,
                                     top_radius=top_radius,
                                     **opts)
        cylinders = cylinder_class(starts, arrow_starts, radius=radius, **opts)

        super().__init__(
            arrowheads, cylinders
        )

class X3DCappedCylinder(X3DGroup):

    cap_class = X3DSphere
    cylinder_class = X3DCylinder

    def __init__(self,
                 starts, ends,
                 radius=1,
                 cylinder_class=None,
                 cap_class=None,
                 cap_offset=0,
                 use_caps=(True, True),
                 **opts):
        if cap_class is None:
            cap_class = self.cap_class
        if cylinder_class is None:
            cylinder_class = self.cylinder_class

        if cap_offset != 0:
            raise NotImplementedError("cap offseting not supported yet")

        if use_caps is True: use_caps = [True, True]
        elif use_caps is False: use_caps = [False, False]

        cap_list = []
        if use_caps[0]:
            cap_list.append(cap_class(starts, radius=radius, **opts))
        if use_caps[1]:
            cap_list.append(cap_class(ends, radius=radius, **opts))
        cylinders = cylinder_class(starts, ends, radius=radius, **opts)

        super().__init__(
            *cap_list, cylinders
        )

class X3DText(X3DGeometryGroup):
    tag_class = X3DHTML.Text

    def __init__(self, *args, billboard=True, billboard_opts=None, **opts):
        if billboard:
            if billboard_opts is None:
                billboard_opts = {'axisOfRotation':'0 0 0'}
            self.wrapper_class = lambda *x,**y:X3DHTML.Billboard(X3DHTML.Shape(*x, **y), **billboard_opts)
        super().__init__(*args, **opts)
    def create_tag_object(self, font_style=None, **core_opts):
        body = []
        if font_style is not None:
            body.append(font_style)
        return self.tag_class(body, **core_opts)
    def prep_geometry_opts(self, centers, text, font_style=None, rotation=None, normal=None, **opts):
        centers = self.prep_vecs(centers)
        rotation = self.prep_vecs(rotation, len(centers))
        normal = self.prep_vecs(normal, len(centers))
        text = self.prep_const(text, len(centers))
        font_style = [
            X3DHTML.FontStyle(**fs)
                if fs is not None else
            None
            for fs in self.prep_const(font_style, len(centers))
        ]
        return [
            {"translation": c, 'string': t, 'length': len(t), 'font_style':fs, 'rotation':r, 'normal':n}
            for c, t, fs, r, n in zip(centers, text, font_style, rotation, normal)
        ]

class X3DTorus(X3DGeometryGroup):
    tag_class = X3DHTML.Torus

    def prep_geometry_opts(self, centers, radius=1, inner_radius=None,
                           normal=None, rotation=None, scale=None,
                           angle=None,
                           **opts):
        if angle is not None and angle < 0:
            if rotation is None:
                rotation = [0, 0, 1, 0]
            if isinstance(rotation, str):
                rotation = np.array(rotation.split()).astype(float)
            rotation = list(rotation[:3]) + [rotation[3] + angle]
            angle = abs(angle)
        centers = self.prep_vecs(centers)
        normal = self.prep_vecs(normal, centers.shape[0])
        rotation = self.prep_vecs(rotation, centers.shape[0])
        scale = self.prep_vecs(scale, centers.shape[0])
        radius = self.prep_const(radius, centers.shape[0])
        inner_radius = self.prep_const(inner_radius, centers.shape[0])
        angle = self.prep_const(angle, centers.shape[0])

        return [
            {
                "translation": s, "normal": n, "outerRadius": r,
                "innerRadius": i, 'rotation': rot, 'scale': sc,
                'angle': ang,
                **opts}
            for s, n, rot, sc, r, i, ang in zip(
                centers, normal,
                rotation, scale,
                radius, inner_radius, angle
            )
        ]

class X3DCoordinatesWrapper(X3DGeometryGroup):
    tag_class: X3DHTML.X3DElement
    def create_tag_object(self, *, point, color=None, **etc):
        body = [X3DCoordinate(point).to_x3d()]
        if color is not None:
            body.append(X3DColor(color).to_x3d())
        return self.tag_class(body, **etc)
    def prep_geometry_opts(self, point, **etc):
        return [
            dict({"translation":"0,0,0", "point":point}, **etc)
        ]

class X3DIndexedCoordinatesWrapper(X3DCoordinatesWrapper):
    def prep_geometry_opts(self, point, indices, vertex_colors=None, **etc):
        base_dict = dict(
                {
                    'index': " ".join(np.asanyarray(indices).flatten().astype(int).astype(str))
                },
                **super().prep_geometry_opts(point, **etc)[0]
            )
        if vertex_colors is not None:
            base_dict['colorPerVertex'] = True
            base_dict['color'] = vertex_colors
        return [base_dict]

class X3DGeometry2DGroup(X3DGeometryGroup):
    def prep_geometry_opts(self, center, **etc):
        # raise Exception(crosses)
        # a = np.concatenate([crosses, [angs]])
        # a, _ = self.get_rotation(crosses, up_vector=up_vector)
        return [
            {
                "translation":center,
                 **etc
            }
        ]

class X3DRectangle2D(X3DGeometry2DGroup):
    tag_class = X3DHTML.Rectangle2D
class X3DCircle2D(X3DGeometry2DGroup):
    tag_class = X3DHTML.Circle2D
    def prep_geometry_opts(self, centers, radius=1,
                           normal=None, rotation=None, scale=None,
                           angle=None,
                           **opts):
        centers = self.prep_vecs(centers)
        normal = self.prep_vecs(normal, centers.shape[0])
        if angle is not None and angle < 0:
            if rotation is None:
                rotation = [0, 0, 1, 0]
            if isinstance(rotation, str):
                rotation = np.array(rotation.split()).astype(float)
            rotation = list(rotation[:3]) + [rotation[3] + angle]
            angle = abs(angle)
        rotation = self.prep_vecs(rotation, centers.shape[0])
        scale = self.prep_vecs(scale, centers.shape[0])
        radius = self.prep_const(radius, centers.shape[0])
        angle = self.prep_const(angle, centers.shape[0])

        return [
            {
                "translation": s, "normal": n, "radius": r,
                'rotation': rot, 'scale': sc,
                'angle': ang,
                **opts}
            for s, n, rot, sc, r, ang in zip(
                centers, normal,
                rotation, scale,
                radius, angle
            )
        ]

class X3DDisk2D(X3DGeometry2DGroup):
    tag_class = X3DHTML.Disk2D

    def prep_geometry_opts(self, centers, radius=1, inner_radius=None,
                           normal=None, rotation=None, scale=None,
                           angle=None,
                           **opts):
        centers = self.prep_vecs(centers)
        normal = self.prep_vecs(normal, centers.shape[0])
        if angle is not None and angle < 0:
            if rotation is None:
                rotation = [0, 0, 1, 0]
            if isinstance(rotation, str):
                rotation = np.array(rotation.split()).astype(float)
            rotation = list(rotation[:3]) + [rotation[3] + angle]
            angle = abs(angle)
        rotation = self.prep_vecs(rotation, centers.shape[0])
        scale = self.prep_vecs(scale, centers.shape[0])
        radius = self.prep_const(radius, centers.shape[0])
        inner_radius = self.prep_const(inner_radius, centers.shape[0])
        angle = self.prep_const(angle, centers.shape[0])

        return [
            {
                "translation": s, "normal": n, "outerRadius": r,
                "innerRadius": i, 'rotation': rot, 'scale': sc,
                'angle': ang,
                **opts}
            for s, n, rot, sc, r, i, ang in zip(
                centers, normal,
                rotation, scale,
                radius, inner_radius, angle
            )
        ]
class X3DPolyline2D(X3DGeometry2DGroup):
    tag_class = X3DHTML.Polyline2D

class X3DPointSet(X3DCoordinatesWrapper):
    tag_class = X3DHTML.PointSet
class X3DLine(X3DCoordinatesWrapper):
    tag_class = X3DHTML.LineSet
class X3DTriangleSet(X3DCoordinatesWrapper):
    tag_class = X3DHTML.TriangleSet
class X3DIndexedTriangleSet(X3DIndexedCoordinatesWrapper):
    tag_class = X3DHTML.IndexedTriangleSet
class X3DIndexedLineSet(X3DIndexedCoordinatesWrapper):
    tag_class = X3DHTML.IndexedLineSet
    def prep_geometry_opts(self, point, indices, **etc):
        opts = super().prep_geometry_opts(point, indices, **etc)
        opts[0]['coordIndex'] = opts[0].pop('index')
        return opts
class X3DIndexedQuadSet(X3DIndexedCoordinatesWrapper):
    tag_class = X3DHTML.IndexedQuadSet
class X3DIndexedFaceSet(X3DIndexedCoordinatesWrapper):
    tag_class = X3DHTML.IndexedFaceSet

class X3DGenericAnimator(X3DGroup):
    def __init__(self, *animation_data, id=None, animation_duration=2, running=True, slider=False,
                 **opts):
        self.uuid = str(uuid.uuid4())
        if id is None:
            id = f"animation-{self.uuid}"

        animated_objects, attributes, nframes = self.get_animation_objects(animation_data, id)
        elements = []
        if slider:
            elements.append(
                JHTML.Input(type="range", value="0", min="0", max=f"{nframes}", step="1", cls="slider",
                            oninput=f"""document.getElementById("{id}").setAttribute("whichChoice", this.value)""")
            )
        elements.extend(animated_objects)
        # if running:
        elements.extend(
            self.build_animator_group(attributes,
                                      nframes=nframes,
                                      animation_duration=animation_duration,
                                      running=running,
                                      uuid=self.uuid
                                      )
        )

        super().__init__(elements, id=id, **opts)

    @classmethod
    @abc.abstractmethod
    def get_animation_objects(cls, animation_data, id) -> tuple[list[X3DObject], dict, int]:
        ...

    @classmethod
    def build_animator_group(cls, attribute_sets, nframes, *, uuid, running=True, animation_duration=2):
        key_frames = np.linspace(0, 1, nframes+1)[:-1]
        return [
            X3DHTML.TimeSensor(id=f'animation-clock-{uuid}', cycleInterval=animation_duration, loop=True,
                               enabled=running),
            X3DHTML.IntegerSequencer(id=f'animation-indexer-{uuid}',
                                     key=key_frames,
                                     keyValue=np.arange(nframes)),
            X3DHTML.Route(
                fromField='fraction_changed', fromNode=f'animation-clock-{uuid}',
                toField='set_fraction', toNode=f'animation-indexer-{uuid}'
            )
        ] + sum((
            cls.create_animation_control(name, uuid=uuid, **opts)
            for attributes in attribute_sets
            for name,opts in attributes.items()),
            []
        )

    @classmethod
    def resolve_control_type(cls, name, values):
        if np.all(values == np.arange(1, len(values)+1)):
            return 'indexed'
        else:
            return 'interpolated'

    interpolator_map = {
        ('color', 'glow'): X3DHTML.ColorInterpolator,
        ('position', 'translation'): X3DHTML.PositionInterpolator,
        ('point', 'coordinate'): X3DHTML.CoordinateInterpolator,
        ('rotation',): X3DHTML.OrientationInterpolator
    }
    @classmethod
    def resolve_interpolator_type(cls, name, values):
        nl = name.lower()
        for k,t in cls.interpolator_map.items():
            if any(kl in nl for kl in k):
                interp_type = t
                if issubclass(interp_type, X3DHTML.PositionInterpolator):
                    values = np.asanyarray(values)
                    if values.ndim == 1:
                        nframes = len(values) // 3
                    else:
                        nframes = len(values)
                elif issubclass(interp_type, X3DHTML.OrientationInterpolator):
                    values = np.asanyarray(values)
                    if values.ndim == 1:
                        nframes = len(values) // 4
                    else:
                        nframes = len(values)
                else:
                    nframes = len(values)
                break
        else:
            values = np.asanyarray(values)
            nframes = len(values)
            if np.issubdtype(values.dtype, np.dtype(str)):
                interp_type = X3DHTML.ColorInterpolator
            elif values.ndim == 1:
                interp_type = X3DHTML.ScalarInterpolator
            elif values.ndim == 2 and values.shape[-1] == 3:
                interp_type = X3DHTML.PositionInterpolator
            elif values.ndim == 3 and values.shape[-1] == 3:
                interp_type = X3DHTML.CoordinateInterpolator
            elif values.ndim == 2 and values.shape[-1] == 4:
                interp_type = X3DHTML.OrientationInterpolator
            else:
                raise ValueError(f"interpolator can't be determined for property {name} with shape {values.shape}")

        return interp_type, nframes

    @classmethod
    def create_animation_control(cls, name, *, id, uuid, type=None, values=None, interpolator_type=None):
        if values is None:
            if type is None:
                raise ValueError("`values` or `type` must be passed")
            return [
                X3DHTML.Route(
                    fromField='value_changed' if type == 'indexed' else 'fraction_changed',
                    fromNode=f'animation-indexer-{uuid}' if type == 'indexed' else f'animation-clock-{uuid}',
                    toField=name, toNode=id
                )
            ]
        else:
            # if type is None:
            #     type = self.resolve_control_type(name, values)
            interpolator_type, nframes = cls.resolve_interpolator_type(name, values)
            if nput.is_numeric_array_like(values):
                values = " ".join(np.round(values, 4).flatten().astype(str))
            key_frames = np.linspace(0, 1, nframes)
            interp_obj = (
                interpolator_type(key=key_frames, keyValue=values, id=id+"-interpolator-"+cls.get_new_id())
                    if issubclass(interpolator_type, X3DHTML.X3DElement) else
                interpolator_type
            )
            if hasattr(interp_obj, 'id'):
                interp_id = interp_obj.id
            else:
                interp_id = interp_obj['id']
            return [
                interp_obj,
                X3DHTML.Route(
                    fromField='fraction_changed',
                    fromNode=f'animation-clock-{uuid}',
                    toField='set_fraction', toNode=interp_id
                ),
                X3DHTML.Route(
                    fromField='value_changed',
                    fromNode=interp_id,
                    toField='set_'+name, toNode=id
                )
            ]

class X3DListAnimator(X3DGenericAnimator):
    @classmethod
    def get_animation_objects(self, frames, id):
        anim_frames = X3DSwitch(
                *frames,
                id=id+"-switch",
                whichChoice="0"
            )
        nframes = len(anim_frames.children)
        attributes = [
            {
                'whichChoice':{'type':'indexed', 'id':id+"-switch"}
            }
        ]
        return [anim_frames], attributes, nframes

class X3DInterpolatingAnimator(X3DGenericAnimator):
    @classmethod
    def get_animation_objects(cls, object_attr_sets:dict[X3DObject, dict], id):
        if isinstance(object_attr_sets, tuple) and len(object_attr_sets) == 1:
            object_attr_sets = object_attr_sets[0]
        objects = []
        att_set = []
        nframes = None
        for i,(obj,attrs) in enumerate(object_attr_sets.items()):
            objects.append(obj)
            at_list = {}
            for a,v in attrs.items():
                if nframes is None:
                    nframes = len(v)
                else:
                    nf = len(v)
                    if nframes != nf:
                        raise ValueError(f"attribute {a} has different number of frames {nf} than other attributes {nframes}")
                prop = obj.resolve_prop_attr(a)
                node = obj.get_prop_node_id(a)
                at_list[prop] = {
                    'id':node,
                    'values':v
                }
            att_set.append(at_list)

        return objects, att_set, nframes

    @classmethod
    def frame_diffs(cls, ref:X3DObject|X3DHTML.X3DElement, test:X3DObject|X3DHTML.X3DElement, *rest:X3DObject|X3DHTML.X3DElement):
        statics = []
        changes = {}
        queue = collections.deque([(ref, test) + rest])
        # avoid bugs
        del ref
        del test
        while queue:
            trees = queue.pop()
            left = trees[0]
            rights = trees[1:]
            if isinstance(left, X3DObject):
                left_kids = left.get_children()
                right_kids = [right.get_children() for right in rights]
            else:
                left_kids = left.elems
                right_kids = [right.elems for right in rights]

            l = len(left_kids)
            for right,rk in zip(rights, right_kids):
                if l != len(rk):
                    raise ValueError(f"tree nodes have different numbers of children {l} vs {len(rk)} for  {left} and {right}")

            if l > 0:
                all_kids = [left_kids] + right_kids
                queue.extend(zip(*all_kids))

            if isinstance(left, X3DObject):
                left_attrs = left.get_interpolated_attributes()
                right_attrs = [right.get_interpolated_attributes() for right in rights]
            else:
                left_attrs = left.attrs
                right_attrs = [right.attrs for right in rights]

            attr_diffs = {}
            all_keys = {
                k
                for a in [left_attrs] + right_attrs
                for k in a.keys()
            } - {'id'}
            for k in all_keys:
                v = left_attrs.get(k)
                vals = [v]
                diffed = False
                for a in right_attrs:
                    v2 = a.get(k)
                    vals.append(v2)
                    if not diffed:
                        diffed = (
                                v is None and v2 is not None
                                or v2 is None and v is None
                        )
                        if not diffed:
                            if nput.is_numeric_array_like(v):
                                if not nput.is_numeric_array_like(v2):
                                    diffed = True
                                else:
                                    diffed = not np.allclose(v, v2)
                            elif nput.is_numeric_array_like(2):
                                diffed = True
                            else:
                                diffed = v2 != v
                if diffed:
                    attr_diffs[k] = vals
            if len(attr_diffs) > 0:
                changes[left] = attr_diffs
            elif l == 0:
                statics.append(left)
        return statics, changes

    @classmethod
    def from_frames(cls, frames:list[X3DObject|X3DHTML.X3DElement], **opts):
        static_objects, interpolated_objects = cls.frame_diffs(*frames)
        if len(interpolated_objects) == 0:
            return frames[0]
        else:
            anim = X3DInterpolatingAnimator(interpolated_objects, **opts)
            if len(static_objects) > 0:
                anim = X3DGroup(static_objects + [anim])
            return anim




"""
Provides various types of plots and plotting utilities
"""
import collections

from .Graphics import Graphics, Graphics3D, GraphicsGrid
from .Backends import GraphicsAxes, GraphicsFigure
from .. import Devutils as dev
import numpy as np
import itertools
# import matplotlib.figure
# import matplotlib.axes

__all__ = [
    "Plot", "DataPlot", "ArrayPlot", "TensorPlot",
    "Plot2D", "ListPlot2D",
    "Plot3D", "ListPlot3D",
    "CompositePlot",
    "resolve_plotter",
    "plot_generic",
    "plot_multi"
]

######################################################################################################
#
#                                    'adaptive' function sampling
#
#
# region function application
def _apply_f(f, grid):
    try:
        vals = f(grid)
    except:
        vals = np.vectorize(f)(grid)

    return vals

def _semi_adaptive_sample_func(f, xmin, xmax, npts=150, max_refines=10, der_cut=10**5):

    refines = 0
    der_good = False

    try:
        from ..Zachary import finite_difference
    except ImportError:
        pass
    else:
        while refines < max_refines and not der_good:
            grid = np.linspace(xmin, xmax, npts)
            vals = _apply_f(f, grid)
            ders = finite_difference(grid, vals, 1, 5)
            der_good = not np.any(abs(ders) > der_cut)
            npts *= 2
            refines += 1

    if refines == 0:  # edge case
        grid = np.linspace(xmin, xmax, npts)
        vals = _apply_f(f, grid)

    return grid, vals, npts, refines

def _semi_adaptive_sample_func2(f, xmin, xmax, ymin, ymax, npts=15, max_refines=10, der_cut=10**5):
    from ..Zachary import finite_difference

    refines = 0
    der_good = False
    try:
        from ..Zachary import finite_difference
    except ImportError:
        pass
    else:
        while refines < max_refines and not der_good:
            grid = np.array(np.meshgrid(np.linspace(xmin, xmax, npts), np.linspace(ymin, ymax, npts))).T
            vals = _apply_f(f, grid)
            ders = finite_difference(grid, vals, (1, 1), (5, 5))
            der_good = not np.any(abs(ders) > der_cut)
            npts *= 2
            refines += 1

    if refines == 0:  # edge case
        grid = np.linspace(xmin, xmax, npts)
        vals = _apply_f(f, grid)

    return grid, vals, npts, refines
# endregion


######################################################################################################
#
#                                    Plot data methods
#
#
# region plot data prep
def _interp2DData(gpts, **opts):
    from scipy.interpolate import griddata

    x = np.sort(gpts[:, 0])
    y = np.sort(gpts[:, 1])

    xmin = np.min(x); xmax = np.max(x)

    xdiffs = np.abs(np.diff(x)); xh = np.min(xdiffs[np.nonzero(xdiffs)])
    ymin = np.min(y); ymax = np.max(y)
    ydiffs = np.abs(np.diff(y)); yh = np.min(xdiffs[np.nonzero(ydiffs)])

    num_x = (xmin - xmax) / xh
    if 5 > num_x or num_x < 10000:  # don't want to get too wild
        num_x = 100  # okay but let's get a little wild
    num_y = (ymin - ymax) / yh
    if 5 > num_y or num_y < 10000:  # don't want to get too wild
        num_y = 100  # okay but let's get a little wild

    # import sys
    # print(num_x, num_y, file = sys.stderr)

    xmesh = np.linspace(xmin, xmax, num_x); ymesh = np.linspace(ymin, ymax, num_y)
    xmesh, ymesh = np.meshgrid(xmesh, ymesh)
    mesh = np.array((xmesh, ymesh)).T
    vals = griddata(gpts[:, (0, 1)], gpts[:, 2], mesh, **opts)

    return xmesh, ymesh, vals.T

def _get_2D_plotdata(func, xrange):
    if hasattr(func, 'subs'):
        from sympy import lambdify
        sym, xrange = xrange
        xrange = np.arange(*xrange)
        fvalues = lambdify([sym], func)(xrange)
    elif not callable(func):
        fvalues = xrange
        xrange = func
    else:
        if len(xrange) == 3 and abs(xrange[2]) < abs(xrange[1] - xrange[0]):
            xrange = np.arange(*xrange)
            fvalues = _apply_f(func, xrange)
        elif len(xrange) > 2:
            fvalues = _apply_f(func, xrange)
        else:
            res = _semi_adaptive_sample_func(func, *xrange)
            xrange = res[0]
            fvalues = res[1]
    return xrange, fvalues

def _get_3D_plotdata(func, xrange, yrange):
    if not callable(func):
        fvalues = yrange
        yrange = xrange
        xrange = func
    else:
        if len(xrange) == 3 and abs(xrange[2]) < abs(xrange[1] - xrange[0]):
            xrange = np.arange(*xrange)
            yrange = np.arange(*yrange)
            xrange, yrange = mesh = np.meshgrid(xrange, yrange)
            fvalues = _apply_f(func, mesh)
        elif len(xrange) > 2 and len(yrange) > 2:
            xrange, yrange = mesh = np.meshgrid(xrange, yrange)
            fvalues = _apply_f(func, mesh)
        else:
            res = _semi_adaptive_sample_func2(func, *xrange, *yrange)
            xrange, yrange = res[0].T
            fvalues = res[1]

    return xrange, yrange, fvalues
# endregion

######################################################################################################
#
#                                    Unified PlotBase class
#
#
# Never implemented this

######################################################################################################
#
#                                    2D Plots on 2D Axes
#
#
class Plot(Graphics):
    """
    The base plotting class to interface into matplotlib or (someday 3D) VTK.
    In the future hopefully we'll be able to make a general-purpose `PlottingBackend` class that doesn't need to be `matplotlib` .
    Builds off of the `Graphics` class to make a unified and convenient interface to generating plots.
    Some sophisticated legwork unfortunately has to be done vis-a-vis tracking constructed lines and other plotting artefacts,
    since `matplotlib` is designed to infuriate.
    """

    line_params = {
        "linewidth", "linestyle", "color", "marker", "markersize",
        "markeredgewidth", "markeredgecolor", "markerfacecolor", "markerfacecoloralt",
        "fillstyle", "antialiased", "dash_capstyle", "solid_capstyle",
        "dash_joinstyle", "solid_joinstyle", "pickradius", "drawstyle", "markevery",
        'gid', "zorder", 'label'
    }
    patch_parms = {
        "agg_filter", "alpha", "animated", "antialiased", "capstyle",
        "clip_box", "clip_on", "clip_path", "color", "edgecolor", "facecolor",
        "figure", "fill", "gid", "hatch", "in_layout", "joinstyle",
         "label", "linestyle", "linewidth", "path_effects",
        "picker", "rasterized", "sketch_params", "snap",
        "transform", "url", "visible", "zorder"
    }

    opt_keys = Graphics.opt_keys | {"plot_style"}

    default_plot_style = {}
    style_mapping = {"format":"fmt"}
    known_styles = {"fmt"} | line_params
    method = "plot"
    def __init__(self,
                 *params,
                 method=None,
                 figure=None, axes=None, subplot_kw=None,
                 plot_style=None, theme=None,
                 **opts
                 ):
        """
        :param params: _empty_ or _x_, _y_ arrays or _function_, _xrange_
        :type params:
        :param plot_style: the plot styling options to be fed into the plot method
        :type plot_style: dict | None
        :param method: the method name as a string or functional form of the method to plot
        :type method: str | function
        :param figure: the Graphics object on which to plot (None means make a new one)
        :type figure: Graphics | None
        :param axes: the axes on which to plot (used in constructing a Graphics, None means make a new one)
        :type axes: None
        :param subplot_kw: the keywords to pass on when initializing the plot
        :type subplot_kw: dict | None
        :param colorbar: whether to use a colorbar or what options to pass to the colorbar
        :type colorbar: None | bool | dict
        :param opts: options to be fed in when initializing the Graphics
        :type opts:
        """

        self.graphics = None

        # we're gonna set things up so that we can have delayed evaluation of the plotting.
        # i.e. a Plot can be initialized but then do all its plotting later
        if plot_style is None:
            plot_style = {}
        for k,v in self.style_mapping.items():
            if k in opts:
                opts[v] = opts[k]
                del opts[k]
        for k in self.known_styles:
            if k in opts:
                plot_style[k] = opts[k]
                del opts[k]
        for k in self.default_plot_style:
            if k not in plot_style:
                plot_style[k] = self.default_plot_style[k]
        self._plot_style = plot_style
        self.plot_opts = opts
        self._initialized = False
        self._data = None

        super().__init__(figure=figure, axes=axes, theme=theme, subplot_kw=subplot_kw, **opts)
        self._init_opts['plot_style'] = plot_style
        if method is None:
           method = self.method
        if isinstance(method, str):
            method = self.axes.get_plotter(method)
        self._method = method

        if len(params) > 0:
            self.plot(*params)

    known_keys = Graphics.known_keys | {
        'method',
        'plot_style',
        'insert_default_styles'
    }
    @classmethod
    def filter_options(cls, opts, allowed=None):
        new = {}
        if allowed is None:
            allowed = cls.known_styles | cls.known_keys
        for k in opts.keys() & allowed:
                new[k] = opts[k]
        return new
    def _check_opts(self, opts):
        diff = opts.keys() - (self.known_styles | self.known_keys)
        if len(diff) > 0:
            raise ValueError("unknown options for {}: {}".format(
                type(self).__name__, list(diff)
            ))

    def _initialize(self):
        self._initialized = True
        self.set_options(**self.plot_opts)

    def _get_plot_data(self, func, xrange):
        xrange, fvalues = _get_2D_plotdata(func, xrange)
        return xrange, fvalues

    def _plot_data(self, *data, **plot_style):
        return self._method(*self._get_plot_data(*data), **plot_style)

    def plot(self, *params, insert_default_styles=True, **plot_style):
        """
        Plots a set of data & stores the result
        :return: the graphics that matplotlib made
        :rtype:
        """
        if insert_default_styles:
            plot_style = dict(self.plot_style, **plot_style)
        self._data = (params, plot_style)
        self.graphics = self._plot_data(*params, **plot_style)
        if not self._initialized:
            self._initialize()
        return self.graphics
    @property
    def artists(self):
        if self.graphics is None or isinstance(self.graphics, list):
            return self.graphics
        else:
            return [self.graphics]

    # def copy(self):
    #     return self.change_figure(None)
    def _change_figure(self, new, *init_args, **init_kwargs):
        """Creates a copy of the object with new axes and a new figure

        :return:
        :rtype:
        """
        # print(init_kwargs)
        # print(self._data[0], init_args)
        # print(init_kwargs, self._init_opts)
        return super()._change_figure(new, *self._data[0], *init_args, **init_kwargs)

    def clear(self):
        """
        Removes the plotted data
        """
        for g in self.graphics:
            self.axes.remove(g)
        self.graphics=None
    def restyle(self, **plot_style):
        """
        Replots the data with updated plot styling
        :param plot_style:
        :type plot_style:
        """
        self.clear()
        self.plot(*self.data, **plot_style)

    @property
    def data(self):
        """
        The data that we plotted
        """
        if self._data is None:
            raise ValueError("{} hasn't been plotted in the first place...")
        return self._data[0]
    @property
    def plot_style(self):
        """
        The styling options applied to the plot
        """
        if self._data is None:
            style = self._plot_style
        else:
            style = self._data[1]
        return style

    def add_colorbar(self, graphics=None, norm=None,  **kw):
        """
        Adds a colorbar to the plot
        """
        if self._initialized:
            if graphics is None and norm is None:
                graphics = self.graphics
            return super().add_colorbar(graphics=graphics, **kw)

    def set_graphics_properties(self, *which, **kw):
        if isinstance(self.graphics, tuple):
            for n,g in enumerate(self.graphics):
                if len(which) == 0 or n in which:
                    self.axes.set_graphics_properties(g, **kw)
        else:
            self.axes.set_graphics_properties(self.graphics, **kw)


    @classmethod
    def merge(cls, main, other, *rest, **kwargs):
        return CompositePlot(main, other, *rest, **kwargs)

    plot_classes = {}
    @classmethod
    def resolve_method(cls, mpl_name):
        return cls.plot_classes[mpl_name]
    # @classmethod
    # def merge_plots(cls, *plots, **styles):
    #     ...
    @classmethod
    def register(cls, plot_class):
        if plot_class.method in cls.plot_classes:
            cls.plot_classes[plot_class.__name__] = plot_class
        else:
            cls.plot_classes[plot_class.method] = plot_class
        return plot_class
Plot.register(Plot)

class CompositePlot:
    def __init__(self, main, other, *rest, **kwargs):
        self.kwargs = kwargs
        self.plots = [main, other, *rest]
    def merge(self, **kwargs):
        base = self.plots[0].change_figure(None, **kwargs)
        for p in self.plots[1:]:
            p.change_figure(base)
        return base
    def show(self, interactive=True):
        self._ref = self.merge(interactive=interactive, **self.kwargs)
        # self._ref.pyplot.mpl_connect()
        self._ref.show()
    def _ipython_display_(self):
        self.show()

@Plot.register
class FilledPlot(Plot):
    """
    Inherits from `Plot`.
    Plots a bunch of x values against a bunch of y values using the `scatter` method.
    """
    known_styles = { "where", "interpolate", "step", "data" } | Plot.patch_parms
    method = "fill_between"

@Plot.register
class ScatterPlot(Plot):
    """
    Inherits from `Plot`.
    Plots a bunch of x values against a bunch of y values using the `scatter` method.
    """
    known_styles = { "s", "c", "marker", "cmap", "norm", "vmin", "vmax",
                     "linewidths", "edgecolors", "plotnonfinite", "data"} | Plot.patch_parms
    style_mapping = {"color":"c", "marker_size":"s"}
    method = "scatter"

class ListScatterPlot(ScatterPlot):
    """
    Inherits from `Plot`.
    Plots a bunch of (x, y) points using the `scatter` method.
    """
    def __init__(self, griddata, **opts):
        super().__init__(griddata[:, 0], griddata[:, 1], **opts)

@Plot.register
class ErrorBarPlot(Plot):
    """
    Inherits from `Plot`.
    Plots error bars using the `errorbar` method.
    """
    known_styles = {"yerr", "xerr", "fmt", "ecolor", "elinewidth", "capsize", "barsabove",
    "lolims", "uplims", "xlolims", "xuplims", "errorevery", "capthick", "data"} | Plot.known_styles
    method = "errorbar"
class ListErrorBarPlot(ErrorBarPlot):
    """A Plot that pulls the errorbar data from a list"""
    def __init__(self, griddata, **opts):
        super().__init__(griddata[:, 0], griddata[:, 1], **opts)

@Plot.register
class StickPlot(Plot):
    """A Plot object that plots sticks"""

    default_plot_style = {'basefmt': " ", 'markerfmt': " "}
    known_styles = {
                       "linefmt", "markerfmt", "basefmt", "bottom", "label", "orientation", "data", 'color',
                       'line_style'
                   } | Plot.known_styles
    method = "stem"
    def plot(self, *params, insert_default_styles=True, **plot_style):
        """
        Plots a set of data | stores the result
        :return: the graphics that matplotlib made
        :rtype:
        """
        if insert_default_styles:
            plot_style = dict(self.plot_style, **plot_style)
        # plot_style = dict(self.plot_style, **plot_style)
        if 'linewidth' in plot_style:
            lw = plot_style['linewidth']
            del plot_style['linewidth']
        else:
            lw = None
        lc = None
        if 'color' in plot_style:
            if 'linefmt' in plot_style:
                raise ValueError("modifying linefmt not currently supported")
            lc = plot_style['color']
            del plot_style['color']
        linefmt = ''
        ls = plot_style.pop('line_style', None)
        if ls is not None:
            if ls == 'dashed':
                if 'linefmt' in plot_style:
                    raise ValueError("modifying passed linefmt not currently supported")
                linefmt+="--"
            elif ls == 'dotted':
                if 'linefmt' in plot_style:
                    raise ValueError("modifying passed linefmt not currently supported")
                linefmt+="-."
            plot_style['linefmt'] = linefmt
        super().plot(*params, insert_default_styles=False, **plot_style)
        if lw is not None:
            self.set_graphics_properties(1, linewidth=lw)
        if lc is not None:
            self.set_graphics_properties(1, color=lc)
        return self.graphics
class ListStickPlot(StickPlot):
    """A Plot object that plots sticks from a list"""
    def __init__(self, griddata, **opts):
        super().__init__(griddata[:, 0], griddata[:, 1], **opts)

@Plot.register
class DatePlot(Plot):
    method = 'plot_date'
    known_styles = {'fmt', 'tz', 'xdate', 'ydate', 'data'} | Plot.known_styles
@Plot.register
class StepPlot(Plot):
    method = 'step'
    known_styles = {'where', 'data'} | Plot.known_styles
@Plot.register
class LogLogPlot(Plot):
    method = 'loglog'
    # known_styles = {}
@Plot.register
class SemiLogXPlot(Plot):
    method = 'semilogx'
    # known_styles = {}
@Plot.register
class SemilogYPlot(Plot):
    method = 'semilogy'
    # known_styles = {}
@Plot.register
class HorizontalFilledPlot(Plot):
    method = 'fill_betweenx'
    known_styles = {'where', 'step', 'interpolate', 'data'} | Plot.patch_parms
@Plot.register
class BarPlot(Plot):
    method = 'bar'
    known_styles = {'height', 'width', 'bottom', 'align', 'data'} | Plot.patch_parms
@Plot.register
class HorizontalBarPlot(Plot):
    method = 'barh'
    known_styles = {'width', 'height', 'left', 'align'} | Plot.patch_parms
# @Plot.register
# class BarLabelPlot(Plot):
#     method = 'bar_label'
#     known_styles = {'container', 'labels', 'fmt', 'label_type', 'padding'}  | Plot.patch_parms
@Plot.register
class EventPlot(Plot):
    method = 'eventplot'
    known_styles = {'positions', 'orientation', 'lineoffsets', 'linelengths', 'linewidths', 'colors', 'linestyles', 'data'} | Plot.line_params
@Plot.register
class PiePlot(Plot):
    method = 'pie'
    known_styles = {'explode', 'labels', 'colors', 'autopct', 'pctdistance',
                    'shadow', 'labeldistance', 'startangle', 'radius', 'counterclock',
                    'wedgeprops', 'textprops', 'center', 'frame', 'rotatelabels', 'normalize',
                    'data'} | Plot.patch_parms
@Plot.register
class StackPlot(Plot):
    method = 'stackplot'
    known_styles = {'labels', 'colors', 'baseline', 'data'} | Plot.patch_parms
@Plot.register
class BrokenHorizontalBarPlot(Plot):
    method = 'broken_barh'
    known_styles = {'xranges', 'yrange', 'data'} | Plot.patch_parms
@Plot.register
class VerticalLinePlot(Plot):
    """
    Plots a bunch of vertical lines
    """
    known_styles = {'ymin', 'ymax', 'colors', 'linestyles', 'label', 'data'} | Plot.line_params
    method = 'vlines'
    def _get_plot_data(self, x, y=1.0):
        if isinstance(y, (int, float)):
            y = [0, y]
        return (x, y)
    def _plot_data(self, *data, **plot_style):
        x, y = data
        return self._method(x, *y, **plot_style)
@Plot.register
class HorizontalLinePlot(Plot):
    """
    Plots a bunch of vertical lines
    """
    known_styles = {'xmin', 'xmax', 'colors', 'linestyles', 'label', 'data'} | Plot.line_params
    method = 'hlines'
    def _get_plot_data(self, y, x=None):
        if x is not None:
            x, y = y, x
        else:
            x = 1.0
        if isinstance(x, (int, float)):
            x = [0, x]
        return (x, y)
    def _plot_data(self, *data, **plot_style):
        x, y = data
        return self._method(y, *x, **plot_style)
#     known_styles = {'xmin', 'xmax', 'colors', 'linestyles', 'label', 'data'}
@Plot.register
class PolygonPlot(Plot):
    method = 'fill'
    known_styles = {'data'} | Plot.patch_parms

@Plot.register
class AxisHorizontalLinePlot(Plot):
    method = 'axhline'
    known_styles = {'xmin', 'xmax'} | Plot.line_params
@Plot.register
class AxisHorizontalSpanPlot(Plot):
    method = 'axhspan'
    known_styles = {'ymin', 'ymax', 'xmin', 'xmax'} | Plot.patch_parms
@Plot.register
class AxisVerticalLinePlot(Plot):
    method = 'axvline'
    known_styles = {'ymin', 'ymax'} | Plot.line_params
@Plot.register
class AxisVeticalSpanPlot(Plot):
    method = 'axvspan'
    known_styles = {'xmin', 'xmax', 'ymin', 'ymax'} | Plot.patch_parms
@Plot.register
class AxisLinePlot(Plot):
    method = 'axline'
    known_styles = {'xy1', 'xy2', 'slope'} | Plot.line_params

@Plot.register
class StairsPlot(Plot):
    method = 'stairs'
    known_styles = {'values', 'edges', 'orientation', 'baseline', 'fill', 'data'} | Plot.patch_parms

# class ClabelPlot(Plot):
#     method = 'clabel'
#     known_styles = {'CS', 'levels'}


# class AnnotatePlot(Plot):
#     method = 'annotate'
#     known_styles = {'text', 'xy'}
# class TextPlot(Plot):
#     method = 'text'
#     known_styles = {'s', 'fontdict'}
# class TablePlot(Plot):
#     method = 'table'
#     known_styles = {'cellText', 'cellColours', 'cellLoc', 'colWidths', 'rowLabels', 'rowColours', 'rowLoc', 'colLabels', 'colColours', 'colLoc',
#                     'loc', 'bbox', 'edges'}
# class ArrowPlot(Plot):
#     method = 'arrow'
#     known_styles = {'dx', 'dy'}
# class InsetAxesPlot(Plot):
#     method = 'inset_axes'
#     known_styles = {'bounds', 'transform', 'zorder'}
# class IndicateInsetPlot(Plot):
#     method = 'indicate_inset'
#     known_styles = {'bounds', 'inset_ax', 'transform', 'facecolor', 'edgecolor', 'alpha', 'zorder'}
# class IndicateInsetZoomPlot(Plot):
#     method = 'indicate_inset_zoom'
#     known_styles = {'inset_ax'}
# class SecondaryXaxisPlot(Plot):
#     method = 'secondary_xaxis'
#     known_styles = {'location', 'functions'}
# class SecondaryYaxisPlot(Plot):
#     method = 'secondary_yaxis'
#     known_styles = {'location', 'functions'}
# class BarbsPlot(Plot):
#     method = 'barbs'
#     known_styles = {'data'}


######################################################################################################
#
#                                    Pure Data Plots on 2D Axes
#
#
class DataPlot(Plot):
    """
    Makes a 2D plot of arbitrary data using a plot method that handles that data type
    """
    image_params = {'cmap', 'norm', 'aspect', 'interpolation', 'alpha', 'vmin', 'vmax', 'origin',
                    'extent', 'interpolation_stage', 'filternorm', 'filterrad', 'resample', 'url', 'data'}
    def __init__(self,
                 *params,
                 plot_style=None, method=None,
                 figure=None, axes=None, subplot_kw=None,
                 colorbar=None,
                 **opts
                 ):
        """
        :param params: _empty_ or _data_
        :type params:
        :param plot_style: the plot styling options to be fed into the plot method
        :type plot_style: dict | None
        :param method: the method name as a string
        :type method: str
        :param figure: the Graphics object on which to plot (None means make a new one)
        :type figure: Graphics | None
        :param axes: the axes on which to plot (used in constructing a Graphics, None means make a new one)
        :type axes: None
        :param subplot_kw: the keywords to pass on when initializing the plot
        :type subplot_kw: dict | None
        :param colorbar: whether to use a colorbar or what options to pass to the colorbar
        :type colorbar: None | bool | dict
        :param opts: options to be fed in when initializing the Graphics
        :type opts:
        """
        super().__init__(*params,
                         plot_style=plot_style, method=method,
                         colorbar=colorbar, figure=figure,
                         axes=axes, subplot_kw=subplot_kw,
                         **opts
                         )
    def _get_plot_data(self, data):
        return data,

@Plot.register
class HistogramPlot(DataPlot):
    """
    Makes a Histogram of data
    """
    method = 'hist'
    known_styles = {'bins', 'range', 'density', 'weights', 'cumulative',
                    'bottom', 'histtype', 'align', 'orientation', 'rwidth', 'log', 'color',
                    'label', 'stacked', 'data'}

@Plot.register
class HistogramPlot2D(DataPlot):
    """
    Makes a 2D histogram of data
    """
    method = 'hist2d'
    known_styles = {'bins', 'range', 'density', 'weights', 'cmin', 'cmax', 'data'}

@Plot.register
class SpectrogramPlot(DataPlot):
    method = 'specgram'
    known_styles = {'NFFT', 'Fs', 'Fc', 'detrend', 'window', 'noverlap',
                    'cmap', 'xextent', 'pad_to', 'sides', 'scale_by_freq', 'mode', 'scale', 'vmin',
                    'vmax', 'data'} | DataPlot.image_params
@Plot.register
class AutocorrelationPlot(DataPlot):
    method = 'acorr'
    known_styles = {'detrend', 'normed', 'usevlines', 'maxlags', 'linestyle', 'marker', 'data'} | Plot.known_styles
@Plot.register
class AngleSpectrumPlot(DataPlot):
    method = 'angle_spectrum'
    known_styles = {'Fs', 'Fc', 'window', 'pad_to', 'sides', 'data'}  | Plot.patch_parms
@Plot.register
class CoherencePlot(DataPlot):
    method = 'cohere'
    known_styles = {'NFFT', 'Fs', 'Fc', 'noverlap', 'pad_to', 'sides', 'scale_by_freq', 'data'} | Plot.known_styles
@Plot.register
class CrossSpectralDensityPlot(DataPlot):
    method = 'csd'
    known_styles = {'NFFT', 'Fs', 'Fc', 'detrend', 'window', 'noverlap', 'pad_to', 'sides', 'scale_by_freq', 'return_line', 'data'}  | Plot.known_styles
@Plot.register
class MagnitudeSpectrumPlot(DataPlot):
    method = 'magnitude_spectrum'
    known_styles = {'Fs', 'Fc', 'window', 'pad_to', 'sides', 'scale', 'data'} | Plot.known_styles
@Plot.register
class PhaseSpectrumPlot(DataPlot):
    method = 'phase_spectrum'
    known_styles = {'Fs', 'Fc', 'window', 'pad_to', 'sides', 'data'} | Plot.known_styles
@Plot.register
class PowerSpectralDensityPlot(DataPlot):
    method = 'psd'
    known_styles = {'NFFT', 'Fs', 'Fc', 'detrend', 'window', 'noverlap', 'pad_to', 'sides', 'scale_by_freq', 'return_line', 'data'} | Plot.known_styles
@Plot.register
class CrossCorrelationPlot(DataPlot):
    method = 'xcorr'
    known_styles = {'normed', 'usevlines', 'maxlags', 'data', 'linestyle', 'marker'} | Plot.known_styles
@Plot.register
class BoxPlot(DataPlot):
    method = 'boxplot'
    known_styles = {'notch', 'sym', 'vert', 'whis', 'positions', 'widths',
                    'patch_artist', 'bootstrap', 'usermedians', 'conf_intervals', 'meanline',
                    'showmeans', 'showcaps', 'showbox', 'showfliers', 'boxprops', 'labels',
                    'flierprops', 'medianprops', 'meanprops', 'capprops', 'whiskerprops',
                    'manage_ticks', 'autorange', 'zorder', 'data'}
@Plot.register
class ViolinPlot(DataPlot):
    method = 'violinplot'
    known_styles = {'dataset', 'positions', 'vert', 'widths', 'showmeans',
                    'showextrema', 'showmedians', 'quantiles', 'points', 'bw_method', 'data'}
# class ViolinPlot(Plot):
#     method = 'violin'
#     known_styles = {'vpstats', 'positions', 'vert', 'widths', 'showmeans',
#                     'showextrema', 'showmedians'}
@Plot.register
class BoxAndWhiskerPlot(DataPlot):
    method = 'bxp'
    known_styles = {'bxpstats', 'positions', 'widths', 'vert', 'patch_artist',
                    'shownotches', 'showmeans', 'showcaps', 'showbox', 'showfliers', 'boxprops',
                    'whiskerprops', 'flierprops', 'medianprops', 'capprops', 'meanprops',
                    'meanline', 'manage_ticks', 'zorder'}
@Plot.register
class HexagonalHistogramPlot(DataPlot):
    method = 'hexbin'
    known_styles = {'C', 'gridsize', 'bins', 'xscale', 'yscale', 'extent', 'cmap', 'norm', 'vmin',
                    'vmax', 'alpha', 'linewidths', 'edgecolors', 'mincnt', 'marginals', 'data', 'reduce_C_function'} | Plot.patch_parms

@Plot.register
@Plot.register
class QuiverPlot(DataPlot):
    method = 'quiver'
    known_styles = {"units", "angles", "scale", "scale_units", "width", "headwidth", "headlength",
                    "headaxislength", "minshaft", "minlength", "pivot", "color", "data"} | Plot.patch_parms
@Plot.register
class StreamPlot(DataPlot):
    method = 'streamplot'
    known_styles = {'density', 'linewidth', 'color', 'cmap', 'norm',
                    'arrowsize', 'arrowstyle', 'minlength', 'transform', 'zorder', 'start_points',
                    'maxlength', 'integration_direction', 'data'}

@Plot.register
class ArrayPlot(DataPlot):
    """
    Plots an array as an image
    """

    method = 'imshow'
    known_styles = DataPlot.image_params
    def __init__(self, *params,
                 plot_style=None, colorbar=None,
                 figure=None, axes=None, subplot_kw=None,
                 **opts
                 ):
        super().__init__(*params,
                         plot_style=plot_style,
                         colorbar=colorbar, figure=figure,
                         axes=axes, subplot_kw=subplot_kw,
                         **opts
                         )
    def _get_plot_data(self, data):
        if hasattr(data, 'toarray'):
            data = data.toarray()
        return data,
@Plot.register
class MatrixPlot(ArrayPlot):
    method = 'matshow'
@Plot.register
class SparsityPlot(ArrayPlot):
    method = 'spy'
    known_styles = {'precision', 'marker', 'markersize', 'aspect', 'origin'} | ArrayPlot.known_styles

class TensorPlot(GraphicsGrid):
    """
    Plots slices of a tensor as a grid
    """
    def __init__(self, tensor,
                 nrows=None, ncols=None,
                 plot_style=None, colorbar=None,
                 figure=None, axes=None, subplot_kw=None,
                 method='imshow', plot_class=None,
                 **opts
                 ):
        from operator import mul
        from functools import reduce
        tensor_shape = tensor.shape
        total_dim = reduce(mul, tensor_shape[:-2], 1)
        if nrows is None or ncols is None:
            if len(tensor_shape) == 3:
                nrows = 1
                ncols = tensor_shape[0]
            elif len(tensor_shape) == 4:  # best case
                nrows, ncols = tensor_shape[:2]
            else:
                if nrows is not None:
                    ncols = total_dim // nrows
                elif ncols is not None:
                    nrows = total_dim // ncols
                else:
                    ncols = 5
                    nrows = total_dim // ncols
        super().__init__(nrows=nrows, ncols=ncols,
                         figure=figure,
                         axes=axes,
                         subplot_kw=subplot_kw
                         )

        tensor = tensor.reshape((total_dim,) + tensor_shape[-2:])
        if plot_class is None:
            plot_class = lambda data,colorbar=colorbar,**opts:ArrayPlot(data, colorbar=colorbar,**opts)
        for i in range(nrows):
            for j in range(ncols):
                graphics = self.axes[i][j]
                self.axes[i][j] = plot_class(
                    tensor[nrows * i + j],
                    figure=graphics,
                    plot_style=plot_style,
                    method=method,
                    **opts
                )

######################################################################################################
#
#                                    3D Plots on 2D Axes
#
#

class Plot2D(Plot):
    """
    A base class for plots of 3D data but plotted on 2D axes
    """
    known_styles = {"corner_mask", "colors", "alpha", "cmap", "norm", "vmin", "vmax", "origin",
                    "extent", "locator", "extend", "xunits, yunits", "antialiased", "nchunk",
                    "linewidths", "linestyles", "hatches", "data"}
    method='contour'
    def __init__(self, *params,
                 plot_style=None,
                 colorbar=None,
                 figure=None,
                 axes=None,
                 subplot_kw=None,
                 **opts
                 ):
        """
        :param params: either _empty_ or _x_, _y_, _z_ arrays or _function_, _xrange_, _yrange_
        :type params:
        :param plot_style: the plot styling options to be fed into the plot method
        :type plot_style: dict | None
        :param method: the method name as a string
        :type method: str
        :param figure: the Graphics object on which to plot (None means make a new one)
        :type figure: Graphics | None
        :param axes: the axes on which to plot (used in constructing a Graphics, None means make a new one)
        :type axes: None
        :param subplot_kw: the keywords to pass on when initializing the plot
        :type subplot_kw: dict | None
        :param colorbar: whether to use a colorbar or what options to pass to the colorbar
        :type colorbar: None | bool | dict
        :param opts: options to be fed in when initializing the Graphics
        :type opts:
        """
        super().__init__(*params,
                         plot_style=plot_style,
                         colorbar=colorbar, figure=figure,
                         axes=axes, subplot_kw=subplot_kw,
                         **opts
                         )
    def _get_plot_data(self, func, xrange, yrange):
        return _get_3D_plotdata(func, xrange, yrange)
@Plot.register
class ContourPlot(Plot2D):
    method = 'contourf'
    known_styles = {"triangles", "mask", "levels", "colors",
                    "alpha", "cmap", "norm", "vmin", "vmax", "origin", "extent", "locator", "extend",
                    "xunits, yunits", "antialiased"}
@Plot.register
class ContourLinePlot(ContourPlot):
    method = 'contour'
    known_styles = ContourPlot.known_styles | {"linewidths", "linestyles", "negative_linestyles"}
@Plot.register
class DensityPlot(Plot2D):
    method = 'pcolormesh'
    known_styles = {'alpha', 'norm', 'cmap', 'vmin', 'vmax', 'shading', 'antialiased', 'data'} | Plot.patch_parms
@Plot.register
class HeatmapPlot(Plot2D):
    method = 'pcolor'
    known_styles = {'shading', 'alpha', 'norm', 'cmap', 'vmin', 'vmax', 'data'} | ArrayPlot.known_styles
@Plot.register
class TriPlot(Plot2D):
    """A Plot object that plots a triangulation bars"""
    method = 'triplot'
    known_styles = {"triangles", "mask"} | Plot.known_styles
class ListTriPlot(Plot2D):
    """A Plot that pulls the triangulation data from a list"""
    def __init__(self, griddata, **opts):
        super().__init__(griddata[:, 0], griddata[:, 1], **opts)
@Plot.register
class TriDensityPlot(Plot2D):
    method = 'tripcolor'
    known_styles = {'alpha', 'norm', 'cmap', 'vmin', 'vmax', 'shading', 'facecolors'}
@Plot.register
class TriContourLinesPlot(Plot2D):
    method = 'tricontour'
    known_styles = {"triangles", "mask", "levels", "colors",
                    "alpha", "cmap", "norm", "vmin", "vmax", "origin", "extent", "locator", "extend",
                    "xunits, yunits", "antialiased", "linewidths", "linestyles"}
@Plot.register
class TriContourPlot(Plot2D):
    method = 'tricontourf'
    known_styles = {"triangles", "mask", "levels", "colors",
                    "alpha", "cmap", "norm", "vmin", "vmax", "origin", "extent", "locator", "extend",
                    "xunits, yunits", "antialiased", "hatches"}

class ListPlot2D(Plot2D):
    """
    Convenience class that handles the interpolation first
    """
    def __init__(self,
                 *params,
                 plot_style=None,
                 method='contour',
                 colorbar=None,
                 figure=None,
                 axes=None,
                 subplot_kw=None,
                 interpolate=True,
                 **opts
                 ):
        """
        :param params: either _empty_ or and array of (_x_, _y_, _z_) points
        :type params:
        :param plot_style: the plot styling options to be fed into the plot method
        :type plot_style: dict | None
        :param method: the method name as a string
        :type method: str
        :param figure: the Graphics object on which to plot (None means make a new one)
        :type figure: Graphics | None
        :param axes: the axes on which to plot (used in constructing a Graphics, None means make a new one)
        :type axes: None
        :param subplot_kw: the keywords to pass on when initializing the plot
        :type subplot_kw: dict | None
        :param colorbar: whether to use a colorbar or what options to pass to the colorbar
        :type colorbar: None | bool | dict
        :param interpolate: whether to interpolate the data or not
        :type interpolate: bool
        :param opts: options to be fed in when initializing the Graphics
        :type opts:
        """
        self.interpolate = interpolate
        super().__init__(*params,
                         plot_style=plot_style, method=method,
                         colorbar=colorbar, figure=figure,
                         axes=axes, subplot_kw=subplot_kw,
                         **opts
                         )

    def _get_plot_data(self, *griddata, interpolate=None):
        if interpolate is None:
            interpolate = self.interpolate
        if len(griddata) == 3:
            x, y, z = griddata
        elif interpolate:
            x, y, z = _interp2DData(griddata[0])
        else:
            x = griddata[0][:, 0]
            y = griddata[0][:, 1]
            z = griddata[0][:, 2]

        return (x, y, z)

@Plot.register
class ListContourPlot(ContourPlot):
    _get_plot_data = ListPlot2D._get_plot_data
    def __init__(self,*params, interpolate=True, **opts):
        self.interpolate = interpolate
        super().__init__(*params, **opts)
@Plot.register
class ListDensityPlot(DensityPlot):
    _get_plot_data = ListPlot2D._get_plot_data
    def __init__(self,*params, interpolate=True, **opts):
        self.interpolate = interpolate
        super().__init__(*params, **opts)
@Plot.register
class ListTriContourPlot(TriContourPlot):
    _get_plot_data = ListPlot2D._get_plot_data
    def __init__(self,*params, interpolate=True, **opts):
        self.interpolate = interpolate
        super().__init__(*params, **opts)
@Plot.register
class ListTriDensityPlot(TriDensityPlot):
    _get_plot_data = ListPlot2D._get_plot_data
    def __init__(self,*params, interpolate=True, **opts):
        self.interpolate = interpolate
        super().__init__(*params, **opts)


######################################################################################################
#
#                                    3D Plots on 3D Axes
#
#
class Plot3D(Graphics3D):  # basically a mimic of the Plot class but inheriting from Graphics3D
    """A base class for 3D plots"""

    default_plot_style = {}
    style_mapping = {"format": "fmt"}
    known_styles = {"fmt"} | Plot.line_params
    method = 'plot_surface'
    def __init__(self, *params,
                 plot_style=None,
                 method=None, colorbar=None,
                 figure=None, axes=None, subplot_kw=None,
                 **opts
                 ):
        """
        :param params: either _empty_ or _x_, _y_, _z_ arrays or _function_, _xrange_, _yrange_
        :type params:
        :param plot_style: the plot styling options to be fed into the plot method
        :type plot_style: dict | None
        :param method: the method name as a string
        :type method: str
        :param figure: the Graphics object on which to plot (None means make a new one)
        :type figure: Graphics | None
        :param axes: the axes on which to plot (used in constructing a Graphics, None means make a new one)
        :type axes: None
        :param subplot_kw: the keywords to pass on when initializing the plot
        :type subplot_kw: dict | None
        :param colorbar: whether to use a colorbar or what options to pass to the colorbar
        :type colorbar: None | bool | dict
        :param opts: options to be fed in when initializing the Graphics
        :type opts:
        """

        super().__init__(figure=figure, axes=axes, subplot_kw=subplot_kw)
        if method is None:
           method = self.method
        if isinstance(method, str):
            method = getattr(self.axes, method)
        self._method = method

        # we're gonna set things up so that we can have delayed evaluation of the plotting.
        # i.e. a Plot3D can be initialized but then do all its plotting later
        if plot_style is None:
            plot_style = {}
        for k,v in self.style_mapping.items():
            if k in opts:
                opts[v] = opts[k]
                del opts[k]
        for k in self.known_styles:
            if k in opts:
                plot_style[k] = opts[k]
        for k in self.default_plot_style:
            if k not in plot_style:
                plot_style[k] = self.default_plot_style[k]
        self._plot_style = plot_style
        self.plot_style = plot_style
        self.plot_opts = opts
        self._colorbar = colorbar
        self._initialized = False

        if len(params) > 0:
            self.plot(*params)

    def _initialize(self):
        self._initialized = True
        self.set_options(**self.plot_opts)
        if self.colorbar:
            self.add_colorbar()
        elif isinstance(self.colorbar, dict):
            self.add_colorbar(**self.colorbar)

    def _get_plot_data(self, func, xrange, yrange):
        return _get_3D_plotdata(func, xrange, yrange)

    def _plot_data(self, *data, **plot_style):
        return self._method(*self._get_plot_data(*data), **plot_style)

    def plot(self, *params, **plot_style):
        plot_style = dict(self.plot_style, **plot_style)
        self.graphics = self._plot_data(*params, **plot_style)
        if not self._initialized:
            self._initialize()
        return self.graphics
    def add_colorbar(self, **kw):
        if self._initialized:
            fig:GraphicsFigure = self.figure
            ax:GraphicsAxes = self.axes
            return fig.create_colorbar(self.graphics, **kw)
        else:
            self._colorbar = kw
    plot_classes = {}
    @classmethod
    def resolve_method(cls, mpl_name):
        return cls.plot_classes[mpl_name]
    # @classmethod
    # def merge_plots(cls, *plots, **styles):
    #     ...
    @classmethod
    def register(cls, plot_class):
        if plot_class.method in cls.plot_classes:
            cls.plot_classes[plot_class.__name__] = plot_class
        else:
            cls.plot_classes[plot_class.method] = plot_class
        return plot_class

@Plot3D.register
class ScatterPlot3D(Plot3D):
    """
    Creates a ScatterPlot of 3D data
    """
    method = 'scatter'

@Plot3D.register
class WireframePlot3D(Plot3D):
    """
    Creates a Wireframe mesh plot of 3D data
    """
    method = 'plot_wireframe'

@Plot3D.register
class ContourPlot3D(Plot3D):
    """
    Creates a 3D ContourPlot of 3D data
    """
    method = 'contourf'

class ListPlot3D(Plot3D):
    """
    Convenience 3D plotting class that handles the interpolation first
    """
    method = 'contour'
    def __init__(self,
                 *params,
                 plot_style=None,
                 method=None,
                 colorbar=None,
                 figure=None,
                 axes=None,
                 subplot_kw=None,
                 interpolate=True,
                 **opts
                 ):
        """
        :param params: either _empty_ or and array of (_x_, _y_, _z_) points
        :type params:
        :param plot_style: the plot styling options to be fed into the plot method
        :type plot_style: dict | None
        :param method: the method name as a string
        :type method: str
        :param figure: the Graphics object on which to plot (None means make a new one)
        :type figure: Graphics | None
        :param axes: the axes on which to plot (used in constructing a Graphics, None means make a new one)
        :type axes: None
        :param subplot_kw: the keywords to pass on when initializing the plot
        :type subplot_kw: dict | None
        :param colorbar: whether to use a colorbar or what options to pass to the colorbar
        :type colorbar: None | bool | dict
        :param interpolate: whether to interpolate the data or not
        :type interpolate: bool
        :param opts: options to be fed in when initializing the Graphics
        :type opts:
        """
        self.interpolate = interpolate
        super().__init__(*params,
                         plot_style=plot_style, method=method,
                         colorbar=colorbar, figure=figure,
                         axes=axes, subplot_kw=subplot_kw,
                         **opts
                         )

    def _get_plot_data(self, *griddata, interpolate=None):
        if interpolate is None:
            interpolate = self.interpolate
        if len(griddata) == 3:
            x, y, z = griddata
        elif interpolate:
            x, y, z = _interp2DData(griddata[0])
        else:
            x = griddata[0][:, 0]
            y = griddata[0][:, 1]
            z = griddata[0][:, 2]

        return (x, y, z)

@Plot3D.register
class ListTriPlot3D(ListPlot3D):
    """
    Creates a triangulated surface plot in 3D
    """
    method = 'plot_trisurf'
    default_plot_style = {}

def resolve_plotter(tag):
    if isinstance(tag, str):
        plotter = Plot.plot_classes.get(tag,Plot3D.plot_classes.get(tag))
        if plotter is None:
            plotter = Plot.plot_classes.get(tag.lower(), Plot3D.plot_classes.get(tag.lower()))
        if plotter is None:
            for v in itertools.chain(Plot.plot_classes.values(), Plot3D.plot_classes.values()):
                if v.method == tag:
                    plotter = v
                    break
        return plotter
    else:
        return tag
def plot_generic(*, x, type='plot', y=None, z=None, func=None, **kwargs):
    plotter = resolve_plotter(type)
    if plotter is None:
        raise ValueError(f"unknown plot type {plotter}")
    if func is not None:
        if y is not None:
            args = [func, x, y]
        else:
            args = [func, x]
    elif z is not None:
        if y is None:
            raise ValueError("need `x` and `y` values when supplied `z`")
        args = [x, y, z]
    else:
        if y is None:
            raise ValueError("need `x` and `y` values when no `func` supplied")
        args = [x, y]
    return plotter(*args, **kwargs)
plot_spec_schema = {'type':str, 'args':list, 'opts':dict}
def plot_multi(
        *plot_specs: dict,
        figure=None,
        plot_type_styles=None,
        default_type='plot',
        x=None,
        y=None,
        z=None,
        func=None,
        common_settings=None,
        **global_settings
):
    if plot_type_styles is None:
        plot_type_styles = {}
    if common_settings is None:
        common_settings = {}
    for key, val in [
        ['x', x],
        ['y', y],
        ['z', z],
        ['func', func]
    ]:
        if not dev.is_default(val):
            common_settings[key] = val
    for i,f in enumerate(plot_specs):
        f = collections.ChainMap(
            f,
            plot_type_styles.get(f.get('type', default_type), {}),
            common_settings
        )
        f['type'] = f.get('type', default_type)
        f['figure'] = f.get('figure', figure)
        func = f.get('func')
        if not dev.is_list_like(func):
            if i == 0:
                figure = plot_generic(**collections.ChainMap(f, global_settings))
            else:
                _ = plot_generic(**f)  # TDB if I prefer to update the object each iteration or not
        else:
            for j, c in enumerate(func):
                if i == 0 and j == 0:
                    figure = plot_generic(**collections.ChainMap(dict(func=c), f, global_settings))
                    if f['figure'] is None:
                        f['figure'] = figure
                else:
                    _ = plot_generic(**collections.ChainMap(dict(func=c), f))
    return figure

# add classes to __all__
for c in Plot.plot_classes.values():
    if c.__name__ not in __all__:
        __all__.append(c.__name__)
for c in Plot3D.plot_classes.values():
    if c.__name__ not in __all__:
        __all__.append(c.__name__)
del c


import numpy as np
from .. import Numputils as nput


__all__ = [
    "convert_cartesian_to_zmatrix",
    "convert_zmatrix_to_cartesians"
]

def get_dists(points, centers):
    return nput.pts_norms(centers, points)

def get_angles(lefts, centers, rights):
    return nput.pts_angles(lefts, centers, rights, return_crosses=False)

def get_diheds(points, centers, seconds, thirds):
    return nput.pts_dihedrals(points, centers, seconds, thirds)


def tile_order_list(ol, ncoords):
    nol = len(ol)
    ncol = len(ol[0])
    fsteps = ncoords / nol
    steps = int(fsteps)
    if steps != fsteps:
        raise ValueError(
            "Number of coordinates {} and number of specifed elements {} misaligned".format(
                ncoords,
                nol
            )
        )
    # broadcasts a single order spec to be a multiple order spec
    base_tile = np.broadcast_to(ol[:, :4], (steps, nol, 4))
    shift = np.reshape(np.arange(0, ncoords, nol), (steps, 1, 1))
    ol_tiled = base_tile + shift
    # now we add on extra flags
    if ncol > 4:
        flags_tiled = np.broadcast_to(ol[:, 4:], (steps, nol, ncol - 4))
        ol_tiled = np.concatenate([ol_tiled, flags_tiled], axis=-1)
    return np.reshape(ol_tiled, (ncoords, ncol))

def convert_cartesian_to_zmatrix(coords, *, ordering, use_rad=True, return_derivs=None, order=None,
            strip_embedding=False,
            derivative_method='new'):
    """The ordering should be specified like:

    [
        [n1],
        [n2, n1]
        [n3, n1/n2, n1/n2]
        [n4, n1/n2/n3, n1/n2/n3, n1/n2/n3]
        [n5, ...]
        ...
    ]

    :param coords:    array of cartesian coordinates
    :type coords:     np.ndarray
    :param use_rad:   whether to user radians or not
    :type use_rad:    bool
    :param ordering:  optional ordering parameter for the z-matrix
    :type ordering:   None or tuple of ints or tuple of tuple of ints
    :param kw:        ignored key-word arguments
    :type kw:
    :return: z-matrix coords
    :rtype: np.ndarray
    """

    if coords.ndim > 2:
        base_shape = coords.shape
        new_coords = np.reshape(coords, (np.prod(base_shape[:-1]),) + base_shape[-1:])
        new_coords, ops = convert_cartesian_to_zmatrix(new_coords,
                                                       ordering=ordering,
                                                       use_rad=use_rad,
                                                       return_derivs=return_derivs,
                                                       order=order,
                                                       strip_embedding=strip_embedding,
                                                       derivative_method=derivative_method)
        single_coord_shape = (base_shape[-2] - 1, new_coords.shape[-1])
        new_shape = base_shape[:-2] + single_coord_shape
        new_coords = np.reshape(new_coords, new_shape)
        if return_derivs:
            ders = ops['derivs']
            # we assume we get a list of derivatives?
            reshaped_ders = [None] * len(ders)
            for i, v in enumerate(ders):
                single_base_shape = (base_shape[-2], new_coords.shape[-1])
                ders_shape = coords.shape + single_base_shape * i + single_coord_shape
                v = v.reshape(ders_shape)
                reshaped_ders[i] = v
                ops['derivs'] = reshaped_ders
        return new_coords, ops

    else:
        if return_derivs is None and order is not None and order > 0:
            return_derivs = True

        ncoords = len(coords)
        orig_ol = np.asanyarray(ordering)#ZMatrixCoordinates.canonicalize_order_list(ncoords, ordering)
        ol = orig_ol
        nol = len(ol)
        ncol = len(ol[0])
        fsteps = ncoords / nol
        steps = int(round(fsteps))

        # print(">> c2z >> ordering:", ol)

        multiconfig = nol < ncoords
        if multiconfig:
            ol = tile_order_list(ol, ncoords)
            mc_ol = ol.copy()

        # we define an order map that we'll index into to get the new indices for a
        # given coordinate
        om = 1 + np.argsort(ol[:, 0])
        if return_derivs and derivative_method != "old":
            og_ord = None
            if order is None:
                order = 1
            elif isinstance(order, list):
                og_ord = order
                order = max(order)

            if strip_embedding:
                specs = [
                    p
                    for n, (i, j, k, l) in enumerate(ordering[1:])
                    for p in (
                        [(i, j), (i, j, k), (i, j, k, l)]
                        if n > 1 else
                        [(i, j), (i, j, k)]
                        if n > 0 else
                        [(i, j)]
                    )
                ]
                if len(specs) > 3:
                    ix = np.concatenate([[0, 1], np.arange(3, len(specs), 3)])
                    jx = np.concatenate([[2], np.arange(4, len(specs), 3)])
                    kx = np.arange(5, len(specs), 3)
                elif len(specs) > 1:
                    ix = np.array([0, 1])
                    jx = np.array([2])
                    kx = None
                else:
                    ix = np.array([0])
                    jx = None
                    kx = None
            else:
                specs = [
                    p
                    for n, (i, j, k, l) in enumerate(ordering[1:])
                    for p in (
                        [(i, j), (i, j, k), (i, j, k, l)]
                    )
                ]
                ix = np.arange(0, len(specs), 3)
                jx = np.arange(1, len(specs), 3)
                kx = np.arange(2, len(specs), 3)

            crd = coords.reshape((steps, nol, 3))
            base_conv = nput.internal_coordinate_tensors(crd, specs, order=order, angle_ordering='ijk')
            if strip_embedding:
                dists = base_conv[0][..., ix]
                if jx is None:
                    angles = np.zeros((len(base_conv), 1))
                else:
                    angles = np.array([
                        np.concatenate([[0], a])
                        for a in base_conv[0][..., jx]
                    ])
                if kx is None:
                    diheds = np.zeros((len(base_conv), 2))
                else:
                    diheds = np.array([
                        np.concatenate([[0], d])
                        for d in base_conv[0][..., kx]
                    ])
            else:
                dists = base_conv[0][..., ix]
                angles = base_conv[0][..., jx]
                diheds = base_conv[0][..., kx]

            base_derivs = base_conv[1:]
            derivs = []
            nats = nol
            for n, d in enumerate(base_derivs):
                deriv = np.zeros(
                    (steps,)
                    + ((nats) * 3,) * (n + 1)
                    + ((nats - 1) * 3,)
                )
                if strip_embedding:
                    deriv[..., np.arange(0, deriv.shape[-1], 3)] = d[..., ix]
                    if jx is not None:
                        deriv[..., np.arange(1, deriv.shape[-1], 3)] = np.pad(
                            d[..., jx],
                            [[0, 0]] * (d.ndim - 1) + [[1, 0]]
                        )
                    if kx is not None:
                        deriv[..., np.arange(2, deriv.shape[-1], 3)] = np.pad(
                            d[..., kx],
                            [[0, 0]] * (d.ndim - 1) + [[2, 0]]
                        )
                else:
                    deriv[..., np.arange(0, deriv.shape[-1], 3)] = d[..., ix]
                    if jx is not None:
                        deriv[..., np.arange(1, deriv.shape[-1], 3)] = d[..., jx]
                    if kx is not None:
                        deriv[..., np.arange(2, deriv.shape[-1], 3)] = d[..., kx]
                derivs.append(deriv)

            if og_ord is not None:
                derivs = [derivs[o - 1] for o in og_ord]
        else:
            # need to check against the cases of like 1, 2, 3 atom molecules
            # annoying but not hard
            if return_derivs:
                derivs = [
                    np.zeros(coords.shape + (nol - 1, 3)),
                    np.zeros(coords.shape + (nol, 3) + (nol - 1, 3))
                ]
            if not multiconfig:
                ix = ol[1:, 0]
                jx = ol[1:, 1]
                dists = get_dists(coords[ix], coords[jx])
                if return_derivs:
                    _dists, dist_derivs, dist_derivs_2 = nput.dist_deriv(coords, ix, jx, order=2)
                    drang = np.arange(len(ix))
                    derivs[0][ix, :, drang, 0] = dist_derivs[0]
                    derivs[0][jx, :, drang, 0] = dist_derivs[1]

                    for i, x1 in enumerate([ix, jx]):
                        for j, x2 in enumerate([ix, jx]):
                            # print(i, j, x1, x2,
                            #       # dist_derivs_2[i, j][0, 0],
                            #       drang
                            #       )
                            derivs[1][x1, :, x2, :, drang, 0] = dist_derivs_2[i, j]

                if len(ol) > 2:
                    ix = ol[2:, 0]
                    jx = ol[2:, 1]
                    kx = ol[2:, 2]
                    angles = np.concatenate((
                        [0], get_angles(coords[ix], coords[jx], coords[kx])
                    ))
                    if not use_rad:
                        angles = np.rad2deg(angles)
                    if return_derivs:
                        _angles, angle_derivs, angle_derivs_2 = nput.angle_deriv(coords, jx, ix, kx, order=2)
                        drang = 1 + np.arange(len(ix))
                        # print(">>>>", np.max(np.abs(angle_derivs)))
                        derivs[0][jx, :, drang, 1] = angle_derivs[0]
                        derivs[0][ix, :, drang, 1] = angle_derivs[1]
                        derivs[0][kx, :, drang, 1] = angle_derivs[2]

                        for i, x1 in enumerate([jx, ix, kx]):
                            for j, x2 in enumerate([jx, ix, kx]):
                                derivs[1][x1, :, x2, :, drang, 1] = angle_derivs_2[i, j]
                else:
                    angles = np.array([0.])
                if len(ol) > 3:
                    ix = ol[3:, 0]
                    jx = ol[3:, 1]
                    kx = ol[3:, 2]
                    lx = ol[3:, 3]
                    if ol.shape[1] == 5:
                        raise NotImplementedError("psi angles might be unnecessary")
                        ix = ix.copy()
                        jx = jx.copy()
                        kx = kx.copy()
                        lx = lx.copy()
                        fx = ol[3:, 4]
                        swap_pos = np.where(fx == 1)
                        swap_i = ix[swap_pos]
                        swap_j = jx[swap_pos]
                        swap_k = kx[swap_pos]
                        swap_l = lx[swap_pos]
                        ix[swap_pos] = swap_l
                        jx[swap_pos] = swap_i
                        kx[swap_pos] = swap_j
                        lx[swap_pos] = swap_k

                    diheds = np.concatenate((
                        [0, 0],
                        get_diheds(coords[ix], coords[jx], coords[kx], coords[lx])
                    ))
                    if not use_rad:
                        diheds = np.rad2deg(diheds)
                    if return_derivs:
                        _diheds, dihed_derivs, dihed_derivs_2 = nput.dihed_deriv(coords, ix, jx, kx, lx, order=2)
                        drang = 2 + np.arange(len(ix))
                        derivs[0][ix, :, drang, 2] = dihed_derivs[0]
                        derivs[0][jx, :, drang, 2] = dihed_derivs[1]
                        derivs[0][kx, :, drang, 2] = dihed_derivs[2]
                        derivs[0][lx, :, drang, 2] = dihed_derivs[3]

                        for i, x1 in enumerate([ix, jx, kx, lx]):
                            for j, x2 in enumerate([ix, jx, kx, lx]):
                                derivs[1][x1, :, x2, :, drang, 2] = dihed_derivs_2[i, j]
                else:
                    diheds = np.array([0, 0])
                ol = ol[1:]

            else:  # multiconfig

                # we do all of this stuff with masking operations in the multiconfiguration cases
                mask = np.repeat(True, ncoords)
                mask[np.arange(0, ncoords, nol)] = False
                ix = ol[mask, 0]
                jx = ol[mask, 1]
                dists = get_dists(coords[ix], coords[jx])
                if return_derivs:
                    _, dist_derivs, dist_derivs_2 = nput.dist_deriv(coords, ix, jx, order=2)
                    drang = np.arange(nol - 1)
                    nreps = int(len(ix) / (nol - 1))
                    drang = np.broadcast_to(drang[np.newaxis], (nreps,) + drang.shape).flatten()
                    derivs[0][ix, :, drang, 0] = dist_derivs[0]
                    derivs[0][jx, :, drang, 0] = dist_derivs[1]

                    for i, x1 in enumerate([ix, jx]):
                        for j, x2 in enumerate([ix, jx]):
                            derivs[1][x1, :, x2 % nol, :, drang, 0] = dist_derivs_2[i, j]

                if nol > 2:
                    # set up the mask to drop all of the first bits
                    mask[np.arange(1, ncoords, nol)] = False
                    ix = ol[mask, 0]
                    jx = ol[mask, 1]
                    kx = ol[mask, 2]
                    angles = get_angles(coords[ix], coords[jx], coords[kx])
                    angles = np.append(angles, np.zeros(steps))
                    insert_pos = np.arange(0, ncoords - 1 * steps - 1, nol - 2)
                    angles = np.insert(angles, insert_pos, 0)
                    angles = angles[:ncoords - steps]
                    if not use_rad:
                        angles = np.rad2deg(angles)
                    if return_derivs:
                        # we might need to mess with the masks akin to the insert call...
                        _, angle_derivs, angle_derivs_2 = nput.angle_deriv(coords, jx, ix, kx, order=2)
                        drang = 1 + np.arange(nol - 2)
                        nreps = int(len(ix) / (nol - 2))
                        drang = np.broadcast_to(drang[np.newaxis], (nreps,) + drang.shape).flatten()
                        derivs[0][jx, :, drang, 1] = angle_derivs[0]
                        derivs[0][ix, :, drang, 1] = angle_derivs[1]
                        derivs[0][kx, :, drang, 1] = angle_derivs[2]

                        for i, x1 in enumerate([ix, jx, kx]):
                            for j, x2 in enumerate([ix, jx, kx]):
                                derivs[1][x1, :, x2 % nol, :, drang, 0] = angle_derivs_2[i, j]
                else:
                    angles = np.zeros(ncoords - steps)

                if nol > 3:
                    # set up mask to drop all of the second atom bits (wtf it means 'second')
                    mask[np.arange(2, ncoords, nol)] = False
                    ix = ol[mask, 0]
                    jx = ol[mask, 1]
                    kx = ol[mask, 2]
                    lx = ol[mask, 3]
                    if ol.shape[1] == 5:
                        raise ValueError("Unclear if there is a difference between tau and psi")
                        ix = ix.copy()
                        jx = jx.copy()
                        kx = kx.copy()
                        lx = lx.copy()
                        fx = ol[mask, 4]
                        swap_pos = np.where(fx == 1)
                        swap_i = ix[swap_pos]
                        swap_j = jx[swap_pos]
                        swap_k = kx[swap_pos]
                        swap_l = lx[swap_pos]
                        ix[swap_pos] = swap_l
                        jx[swap_pos] = swap_i
                        kx[swap_pos] = swap_j
                        lx[swap_pos] = swap_k
                    # print(ol)

                    diheds = get_diheds(coords[ix], coords[jx], coords[kx], coords[lx])
                    # pad diheds to be the size of ncoords
                    diheds = np.append(diheds, np.zeros(2 * steps))

                    # insert zeros where undefined
                    diheds = np.insert(diheds, np.repeat(np.arange(0, ncoords - 2 * steps - 1, nol - 3), 2), 0)
                    # take only as many as actually used
                    diheds = diheds[:ncoords - steps]
                    if not use_rad:
                        diheds = np.rad2deg(diheds)
                    if return_derivs:
                        # Negative sign because my dihed_deriv code is for slightly different
                        # ordering than expected
                        _, dihed_derivs, dihed_derivs_2 = nput.dihed_deriv(coords, ix, jx, kx, lx, order=2)
                        drang = 2 + np.arange(nol - 3)
                        nreps = int(len(ix) / (nol - 3))
                        drang = np.broadcast_to(drang[np.newaxis], (nreps,) + drang.shape).flatten()
                        derivs[0][ix, :, drang, 2] = dihed_derivs[0]
                        derivs[0][jx, :, drang, 2] = dihed_derivs[1]
                        derivs[0][kx, :, drang, 2] = dihed_derivs[2]
                        derivs[0][lx, :, drang, 2] = dihed_derivs[3]

                        for i, x1 in enumerate([ix, jx, kx, lx]):
                            for j, x2 in enumerate([ix, jx, kx, lx]):
                                derivs[1][x1, :, x2 % nol, :, drang, 0] = dihed_derivs_2[i, j]

                else:
                    diheds = np.zeros(ncoords - steps)

                # after the np.insert calls we have the right number of final elements, but too many
                # ol and om elements and they're generally too large
                # so we need to shift them down and mask out the elements we don't want
                mask = np.repeat(True, ncoords)
                mask[np.arange(0, ncoords, nol)] = False
                ol = np.reshape(ol[mask], (steps, nol - 1, ncol)) - np.reshape(np.arange(steps), (steps, 1, 1))
                ol = np.reshape(ol, (ncoords - steps, ncol))
                om = np.reshape(om[mask], (steps, nol - 1)) - nol * np.reshape(np.arange(steps), (steps, 1)) - 1
                om = np.reshape(om, (ncoords - steps,))

        final_coords = np.array(
            [
                dists, angles, diheds
            ]
        ).T

        if multiconfig:
            # figure out what to use for the axes
            origins = coords[mc_ol[1::nol, 1]]
            x_axes = coords[mc_ol[1::nol, 0]] - origins  # the first displacement vector
            y_axes = coords[mc_ol[
                2::nol, 0]] - origins  # the second displacement vector (just defines the x-y plane, not the real y-axis)
            axes = np.array([x_axes, y_axes]).transpose((1, 0, 2))
        else:
            origins = coords[ol[0, 1]]
            axes = np.array([coords[ol[0, 0]] - origins, coords[ol[1, 0]] - origins])

        ol = orig_ol
        om = om - 1
        if ncol == 5:
            ordering = np.array([
                np.argsort(ol[:, 0]), om[ol[:, 1]], om[ol[:, 2]], om[ol[:, 3]], ol[:, 4]
            ]).T
        else:
            ordering = np.array([
                np.argsort(ol[:, 0]), om[ol[:, 1]], om[ol[:, 2]], om[ol[:, 3]]
            ]).T
        opts = dict(use_rad=use_rad, ordering=ordering, origins=origins, axes=axes)

        # if we're returning derivs, we also need to make sure that they're ordered the same way the other data is...
        if return_derivs:
            opts['derivs'] = derivs  # [:1]

        return final_coords, opts


def convert_zmatrix_to_cartesians(
        coordlist,
        *,
        ordering,
        origins=None,
        axes=None,
        use_rad=True,
        return_derivs=None,
        order=None
):
    """Expects to get a list of configurations
    These will look like:
        [
            [dist, angle, dihedral]
            ...
        ]
    and ordering will be
        [
            [pos, point, line, plane]
            ...
        ]
    **For efficiency it is assumed that all configurations have the same length**

    :param coordlist:
    :type coordlist:
    :param origins:
    :type origins:
    :param axes:
    :type axes:
    :param use_rad:
    :type use_rad:
    :param kw:
    :type kw:
    :param ordering:
    :type ordering:
    :param return_derivs:
    :type return_derivs:
    :return:
    :rtype:
    """
    # TODO: introduce fast derivs back into this coordinate system by extracting "specs" from the ordering
    #      and then partially inverting
    if return_derivs: raise NotImplementedError("analytic derivatives need maintenance")

    # make sure we have the ordering stuff in hand
    ordering = np.asanyarray(ordering)
    coordlist = np.asanyarray(coordlist)

    # if np.min(ordering) > 0:
    #     ordering = ordering - 1
    # dim_diff = coordlist.ndim - ordering.ndim
    # if dim_diff > 0:
    #     missing = coordlist.shape[:dim_diff]
    #     ordering = np.broadcast_to(ordering, missing + ordering.shape )

    sysnum = len(coordlist)
    coordnum = len(coordlist[0])
    if ordering.ndim == 2:
        ordering = np.repeat(ordering[np.newaxis], sysnum, axis=0)

    if ordering.shape[-1] > 3:
        atom_ordering = ordering[:, :, 0]
        ordering = ordering[:, 1:, 1:]
    else:
        atom_ordering = None

    total_points = np.empty((sysnum, coordnum + 1, 3))
    return_deriv_order = order
    if return_deriv_order is None:
        if return_derivs is not True and return_derivs is not False and isinstance(return_derivs, int):
            return_derivs = True
            return_deriv_order = return_derivs
        elif return_derivs:
            return_deriv_order = 2
    if return_derivs is None:
        return_derivs = return_deriv_order is not None

    if return_derivs:
        derivs = [
            None,  # no need to stoare a copy of total_points here...
            np.zeros((sysnum, coordnum, 3, coordnum + 1, 3)),
            np.zeros((sysnum, coordnum, 3, coordnum, 3, coordnum + 1, 3))
        ]

    # first we put the origin whereever the origins are specified
    if origins is None:
        origins = [0, 0, 0]
    origins = np.asarray(origins)
    if len(origins.shape) < 2:
        origins = np.broadcast_to(origins, (sysnum, 3))
    total_points[:, 0] = origins

    # set up the next points by just setting them along the x-axis by default
    if axes is None:
        axes = [1, 0, 0]
    axes = np.asarray(axes)
    if axes.ndim == 1:
        axes = np.array([
            axes,
            [0, 1, 0]
        ])  # np.concatenate((np.random.uniform(low=.5, high=1, size=(2,)), np.zeros((1,)) ))])
    if axes.ndim == 2:
        axes = np.broadcast_to(axes[np.newaxis], (sysnum, 2, 3))
    x_pts = origins + nput.vec_normalize(axes[:, 0])
    y_pts = origins + nput.vec_normalize(axes[:, 1])

    dists = coordlist[:, 0, 0]
    if return_derivs:
        raise NotImplementedError('old Z-matrix derivatives disabled for instabilities')
        der_stuff = cartesian_from_rad_derivatives(origins,
                                                   x_pts, y_pts, dists,
                                                   None, None,
                                                   0,
                                                   np.full((len(dists),), -1, dtype=int),
                                                   np.full((len(dists),), -1, dtype=int),
                                                   np.full((len(dists),), -1, dtype=int),
                                                   derivs,
                                                   order=return_deriv_order
                                                   )
        total_points[:, 1] = der_stuff[0]
        if return_deriv_order > 0:
            derivs[1][np.arange(sysnum), :1, :, 1, :] = der_stuff[1]
        if return_deriv_order > 1:
            derivs[2][np.arange(sysnum), :1, :, :1, :, 1, :] = der_stuff[2]

    else:
        ref_points_1, _ = nput.cartesian_from_rad(origins, x_pts, y_pts, dists, None, None)
        total_points[:, 1] = ref_points_1

    # print(">> z2c >> ordering", ordering[0])

    # iteratively build the rest of the coords with one special cases for n=2
    for i in range(1, coordnum):
        # Get the distances away

        ref_coords1 = ordering[:, i, 0]  # reference atom numbers for first coordinate
        refs1 = total_points[np.arange(sysnum), ref_coords1.astype(int)]  # get the actual reference coordinates
        dists = np.reshape(coordlist[:, i, 0], (sysnum, 1))  # pull the requisite distances

        ref_coords2 = ordering[:, i, 1]  # reference atom numbers for second coordinate
        refs2 = total_points[
            np.arange(sysnum), ref_coords2.astype(int)]  # get the actual reference coordinates for the angle
        angle = coordlist[:, i, 1]  # pull the requisite angle values
        if not use_rad:
            angle = np.deg2rad(angle)

        if i == 1:
            refs3 = y_pts
            dihed = None
            ref_coords3 = np.full((len(dists),), -1, dtype=int)
            psi_flag = False
        else:
            ref_coords3 = ordering[:, i, 2]  # reference atom numbers for dihedral ref coordinate
            refs3 = total_points[
                np.arange(sysnum), ref_coords3.astype(int)]  # get the actual reference coordinates for the dihed
            dihed = coordlist[:, i, 2]  # pull proper dihedral values
            if not use_rad:
                dihed = np.deg2rad(dihed)
            if ordering.shape[-1] == 4:
                raise ValueError("Unclear if there is a difference between tau and psi")
                psi_flag = ordering[:, i, 3] == 1
                # dihed[psi_flag] = -dihed[psi_flag]
            else:
                psi_flag = False

        if return_derivs:
            if ordering.shape[-1] == 4:
                raise NotImplementedError("don't have derivatives for case with psi angles")
            der_stuff = nput.cartesian_from_rad_derivatives(
                refs1, refs2, refs3,
                dists, angle, dihed,
                i,
                ref_coords1,
                ref_coords2,
                ref_coords3,
                derivs,
                order=return_deriv_order
            )
            # crd, d1, d2 = stuff

            total_points[:, i + 1] = der_stuff[0]
            if return_deriv_order > 0:
                derivs[1][np.arange(sysnum), :i + 1, :, i + 1, :] = der_stuff[1]
            if return_deriv_order > 1:
                derivs[2][np.arange(sysnum), :i + 1, :, :i + 1, :, i + 1, :] = der_stuff[2]
        else:
            ref_points_1, _ = nput.cartesian_from_rad(refs1, refs2, refs3, dists, angle, dihed, psi=psi_flag)
            total_points[:, i + 1] = ref_points_1

    if atom_ordering is not None:
        rev_ord = atom_ordering  # np.argsort(atom_ordering, axis=1)
        total_points = total_points[np.arange(len(atom_ordering))[:, np.newaxis], rev_ord]  # wat?

    converter_opts = dict(use_rad=use_rad, ordering=ordering)
    if return_derivs:
        if return_deriv_order > 0:
            converter_opts['derivs'] = derivs[1:][:return_deriv_order]

    return total_points, converter_opts